# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
import json
import os

from botocore.loaders import Loader

from awscli.testutils import (
    BaseAWSCommandParamsTest,
    BaseAWSHelpOutputTest,
    FileCreator,
    create_clidriver,
)

# NOTE: Typically, the functional tests reuse preexisting models. However
# because document types was released without any service model using them,
# we are using through a fake service model to fully test document type
# behavior.
DOCTYPE_MODEL = {
    "version": "2.0",
    "metadata": {
        "apiVersion": "2011-06-15",
        "endpointPrefix": "doctype",
        "protocol": "json",
        "jsonVersion": "1.1",
        "serviceAbbreviation": "AWS DocType",
        "serviceFullName": "AWS DocType",
        "serviceId": "DocType",
        "signatureVersion": "v4",
        "targetPrefix": "DocType",
        "uid": "doctype-2011-06-15",
    },
    "operations": {
        "DescribeResource": {
            "name": "DescribeResource",
            "http": {"method": "POST", "requestUri": "/"},
            "input": {"shape": "DescribeResourceShape"},
            "output": {"shape": "DescribeResourceShape"},
            "errors": [],
            "documentation": "<p>Describes resource.</p>",
        }
    },
    "shapes": {
        "DescribeResourceShape": {
            "type": "structure",
            "members": {
                "DocTypeParam": {"shape": "DocType"},
                "ModeledMixedWithDocTypeParam": {"shape": "Mixed"},
                "ListOfDocTypesParam": {"shape": "ListOfDocTypes"},
                "MapOfDocTypesParam": {"shape": "MapOfDocTypes"},
                "NestedListsOfDocTypesParam": {
                    "shape": "NestedListsOfDocTypes"
                },
            },
        },
        "DocType": {
            "type": "structure",
            "members": {},
            "document": True,
            "documentation": "<p>Document type</p>",
        },
        "ListOfDocTypes": {"type": "list", "member": {"shape": "DocType"}},
        "MapOfDocTypes": {
            "type": "map",
            "key": {"shape": "String"},
            "value": {"shape": "DocType"},
        },
        "NestedListsOfDocTypes": {
            "type": "list",
            "member": {"shape": "ListOfDocTypes"},
        },
        "Mixed": {
            "type": "structure",
            "members": {
                "DocType": {"shape": "DocType"},
                "StringMember": {"shape": "String"},
                "ListOfDocTypes": {"shape": "ListOfDocTypes"},
                "MapOfDocTypes": {"shape": "MapOfDocTypes"},
                "NestedListsOfDocTypes": {"shape": "NestedListsOfDocTypes"},
            },
            "documentation": (
                "<p>Structure with modeled and document type parameter</p>"
            ),
        },
        "String": {"type": "string"},
    },
}


def _add_doctype_service_model(file_creator, session, model=None):
    if model is None:
        model = DOCTYPE_MODEL
    file_creator.create_file(
        os.path.join('doctype', '2011-06-15', 'service-2.json'),
        json.dumps(model),
    )
    data_path = session.get_config_variable('data_path').split(os.pathsep)
    loader = Loader()
    loader.search_paths.extend(data_path + [file_creator.rootdir])
    session.register_component('data_loader', loader)


class TestDocumentTypeIO(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestDocumentTypeIO, self).setUp()
        self.files = FileCreator()
        _add_doctype_service_model(self.files, self.driver.session)

    def tearDown(self):
        super(TestDocumentTypeIO, self).tearDown()
        self.files.remove_all()

    def assert_raises_shorthand_syntax_error(
        self, cmdline, stderr_contains=None
    ):
        _, stderr, _ = self.run_cmd(cmdline, expected_rc=252)
        if stderr_contains:
            self.assertIn(stderr_contains, stderr)

    def nested_doctype_shorthand_error(self, parameter_name, member_name):
        return (
            "Error parsing parameter '%s': Shorthand syntax does not support "
            "document types. Use JSON input for top-level argument to specify "
            "nested parameter: %s" % (parameter_name, member_name)
        )

    def test_can_provide_json_for_doc_type(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--doc-type-param',
            '{"foo":"bar"}',
        ]
        self.assert_params_for_cmd(
            cmdline, params={'DocTypeParam': {"foo": "bar"}}
        )

    def test_can_provide_json_for_doc_type_with_scalar_value(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--doc-type-param',
            '"json-string"',
        ]
        self.assert_params_for_cmd(
            cmdline, params={'DocTypeParam': 'json-string'}
        )

    def test_can_provide_json_for_doc_type_in_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--list-of-doc-types-param',
            '["foo", {"bar": "baz"}, 1, null]',
        ]
        self.assert_params_for_cmd(
            cmdline,
            params={'ListOfDocTypesParam': ["foo", {"bar": "baz"}, 1, None]},
        )

    def test_can_provide_json_for_doc_type_in_map(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--map-of-doc-types-param',
            '{"key1": "foo", "key2": {"bar": "baz"}, "key3": 1}',
        ]
        self.assert_params_for_cmd(
            cmdline,
            params={
                'MapOfDocTypesParam': {
                    "key1": "foo",
                    "key2": {"bar": "baz"},
                    "key3": 1,
                }
            },
        )

    def test_can_provide_json_for_doc_type_in_nested_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--nested-lists-of-doc-types-param',
            '[["foo", {"bar": "baz"}, 1, null]]',
        ]
        self.assert_params_for_cmd(
            cmdline,
            params={
                'NestedListsOfDocTypesParam': [
                    ["foo", {"bar": "baz"}, 1, None]
                ]
            },
        )

    def test_can_provide_json_for_nested_doc_type(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            (
                '{'
                '   "DocType": {"foo": "bar"},'
                '   "ListOfDocTypes": ["foo", {"bar": "baz"}, 1, null],'
                '   "MapOfDocTypes": {'
                '       "key1": "foo", "key2": {"bar": "baz"}, "key3": 1},'
                '   "NestedListsOfDocTypes":[["foo", {"bar": "baz"}, 1, null]]'
                '}'
            ),
        ]
        self.assert_params_for_cmd(
            cmdline,
            params={
                'ModeledMixedWithDocTypeParam': {
                    'DocType': {'foo': 'bar'},
                    'ListOfDocTypes': ["foo", {"bar": "baz"}, 1, None],
                    'MapOfDocTypes': {
                        "key1": "foo",
                        "key2": {"bar": "baz"},
                        "key3": 1,
                    },
                    'NestedListsOfDocTypes': [
                        ["foo", {"bar": "baz"}, 1, None]
                    ],
                }
            },
        )

    def test_shorthand_not_supported_for_doc_type_argument(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--doc-type-param',
            'foo=1',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=(
                "Error parsing parameter '--doc-type-param': Invalid JSON"
            ),
        )

    def test_shorthand_not_supported_for_doc_type_in_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--list-of-doc-types-param',
            'bar,1,null',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=(
                "Error parsing parameter '--list-of-doc-types-param': "
                "Invalid JSON"
            ),
        )

    def test_shorthand_not_supported_for_doc_type_in_map(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--map-of-doc-types-param',
            'key1={foo=bar}',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=(
                "Error parsing parameter '--map-of-doc-types-param': "
                "Invalid JSON"
            ),
        )

    def test_shorthand_not_supported_for_doc_type_in_nested_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--nested-lists-of-doc-types-param',
            '[bar,1,null],[foo,2]',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=(
                "Error parsing parameter '--nested-lists-of-doc-types-param': "
                "Invalid JSON"
            ),
        )

    def test_shorthand_not_supported_for_nested_doc_type(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            'DocType={foo=bar}',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=self.nested_doctype_shorthand_error(
                '--modeled-mixed-with-doc-type-param', member_name='DocType'
            ),
        )

    def test_shorthand_not_supported_for_nested_doc_type_in_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            'ListOfDocTypes=[{foo=bar}]',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=self.nested_doctype_shorthand_error(
                '--modeled-mixed-with-doc-type-param',
                member_name='ListOfDocTypes',
            ),
        )

    def test_shorthand_not_supported_for_nested_doc_type_in_map(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            'MapOfDocTypes={key={foo=bar}}',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=self.nested_doctype_shorthand_error(
                '--modeled-mixed-with-doc-type-param',
                member_name='MapOfDocTypes',
            ),
        )

    def test_shorthand_not_supported_for_nested_doc_type_in_nested_list(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            'NestedListsOfDocTypes=[[{foo=bar}]]',
        ]
        self.assert_raises_shorthand_syntax_error(
            cmdline,
            stderr_contains=self.nested_doctype_shorthand_error(
                '--modeled-mixed-with-doc-type-param',
                member_name='NestedListsOfDocTypes',
            ),
        )

    def test_can_use_shorthand_if_only_modeled_members_used(self):
        cmdline = [
            'doctype',
            'describe-resource',
            '--modeled-mixed-with-doc-type-param',
            'StringMember=str-val',
        ]
        self.assert_params_for_cmd(
            cmdline,
            params={
                'ModeledMixedWithDocTypeParam': {
                    'StringMember': 'str-val',
                }
            },
        )

    def test_can_generate_cli_skeleton(self):
        cmdline = ['doctype', 'describe-resource', '--generate-cli-skeleton']
        stdout, _, _ = self.run_cmd(cmdline)
        skeleton_output = json.loads(stdout)
        self.assertEqual(
            skeleton_output,
            {
                'DocTypeParam': {},
                'ModeledMixedWithDocTypeParam': {
                    'DocType': {},
                    'StringMember': '',
                    'ListOfDocTypes': [{}],
                    'MapOfDocTypes': {'KeyName': {}},
                    'NestedListsOfDocTypes': [[{}]],
                },
                'ListOfDocTypesParam': [{}],
                'MapOfDocTypesParam': {'KeyName': {}},
                'NestedListsOfDocTypesParam': [[{}]],
            },
        )


class TestDocTypesHelp(BaseAWSHelpOutputTest):
    def setUp(self):
        super(TestDocTypesHelp, self).setUp()
        self.files = FileCreator()
        _add_doctype_service_model(self.files, self.driver.session)

    def tearDown(self):
        super(TestDocTypesHelp, self).tearDown()
        self.files.remove_all()

    def run_help(self):
        self.driver.main(['doctype', 'describe-resource', 'help'])

    def filter_params_in_model(self, include_params):
        model = copy.deepcopy(DOCTYPE_MODEL)
        request_members = model['shapes']['DescribeResourceShape']['members']
        for request_member in list(request_members):
            if request_member not in include_params:
                del request_members[request_member]

        self.driver = create_clidriver()
        _add_doctype_service_model(self.files, self.driver.session, model)

    def assert_has_document_value_in_json_syntax(self, parameter):
        self.assert_text_order(
            'JSON Syntax::', '{...}', starting_from=parameter
        )

    def assert_no_shorthand_syntax(self):
        self.assert_not_contains('Shorthand Sytanx::')

    def test_includes_note_about_document_types(self):
        self.run_help()
        self.assert_contains('uses document type values')

    def test_does_not_include_note_if_no_document_types(self):
        self.filter_params_in_model([])
        self.assert_not_contains('uses document type values')

    def test_annotates_as_document_type_for_input(self):
        self.run_help()
        self.assert_contains('``--doc-type-param`` (document)')

    def test_annotates_as_document_type_for_output(self):
        self.run_help()
        self.assert_contains('DocTypeParam -> (document)')

    def test_json_syntax_for_doc_type_argument(self):
        self.filter_params_in_model(include_params=['DocTypeParam'])
        self.run_help()
        self.assert_has_document_value_in_json_syntax(
            parameter='--doc-type-param'
        )

    def test_json_syntax_for_doc_type_in_list(self):
        self.filter_params_in_model(include_params=['ListOfDocTypesParam'])
        self.run_help()
        self.assert_has_document_value_in_json_syntax(
            parameter='--list-of-doc-types-param'
        )

    def test_json_syntax_for_doc_type_in_map(self):
        self.filter_params_in_model(include_params=['MapOfDocTypesParam'])
        self.run_help()
        self.assert_has_document_value_in_json_syntax(
            parameter='--map-of-doc-types-param'
        )

    def test_json_syntax_for_doc_type_in_nested_list(self):
        self.filter_params_in_model(
            include_params=['NestedListsOfDocTypesParam']
        )
        self.run_help()
        self.assert_has_document_value_in_json_syntax(
            parameter='--nested-lists-of-doc-types-param'
        )

    def test_json_syntax_for_nested_doc_type(self):
        self.filter_params_in_model(
            include_params=['ModeledMixedWithDocTypeParam']
        )
        self.run_help()
        self.assert_text_order(
            'JSON Syntax::',
            '"DocType": {...},',
            '"ListOfDocTypes": [',
            '"MapOfDocTypes": {"string": {...}',
            '"NestedListsOfDocTypes": [',
            starting_from='--modeled-mixed-with-doc-type-param',
        )

    def test_shorthand_not_documented_for_doc_type_argument(self):
        self.filter_params_in_model(include_params=['DocTypeParam'])
        self.run_help()
        self.assert_no_shorthand_syntax()

    def test_shorthand_not_documented_for_doc_type_in_list(self):
        self.filter_params_in_model(include_params=['ListOfDocTypesParam'])
        self.run_help()
        self.assert_no_shorthand_syntax()

    def test_shorthand_not_documented_for_doc_type_in_map(self):
        self.filter_params_in_model(include_params=['MapOfDocTypesParam'])
        self.run_help()
        self.assert_no_shorthand_syntax()

    def test_shorthand_not_documented_for_doc_type_in_nested_list(self):
        self.filter_params_in_model(
            include_params=['NestedListsOfDocTypesParam']
        )
        self.run_help()
        self.assert_no_shorthand_syntax()

    def test_documents_shorthand_for_only_modeled_members(self):
        self.filter_params_in_model(
            include_params=['ModeledMixedWithDocTypeParam']
        )
        self.run_help()
        self.assert_text_order(
            'Shorthand Syntax::',
            'StringMember=string',
            starting_from='--modeled-mixed-with-doc-type-param',
        )
        self.assert_not_contains('DocTypeParam=')
        self.assert_not_contains('ListOfDocTypesParam=')
        self.assert_not_contains('MapOfDocTypesParam=')
        self.assert_not_contains('NestedListsOfDocTypesParam=')



#!/usr/bin/env python
# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from awscli.testutils import BaseAWSCommandParamsTest, BaseAWSHelpOutputTest


class TestGetObject(BaseAWSCommandParamsTest):
    prefix = ['s3api', 'select-object-content']

    def setUp(self):
        super(TestGetObject, self).setUp()
        self.parsed_response = {'Payload': self.create_fake_payload()}
        self._tempdir = tempfile.mkdtemp()

    def tearDown(self):
        super(TestGetObject, self).tearDown()
        shutil.rmtree(self._tempdir)

    def create_fake_payload(self):
        yield {'Records': {'Payload': b'a,b,c,d\n'}}
        # These next two events are ignored because they aren't
        # "Records".
        yield {
            'Progress': {
                'Details': {
                    'BytesScanned': 1048576,
                    'BytesProcessed': 37748736,
                }
            }
        }
        yield {'Records': {'Payload': b'e,f,g,h\n'}}
        yield {
            'Stats': {
                'Details': {
                    'BytesProcessed': 62605400,
                    'BytesScanned': 1662276,
                }
            }
        }
        yield {'End': {}}

    def test_can_stream_to_file(self):
        filename = os.path.join(self._tempdir, 'outfile')
        cmdline = self.prefix[::]
        cmdline.extend(['--bucket', 'mybucket'])
        cmdline.extend(['--key', 'mykey'])
        cmdline.extend(['--expression', 'SELECT * FROM S3Object'])
        cmdline.extend(['--expression-type', 'SQL'])
        cmdline.extend(['--request-progress', 'Enabled=True'])
        cmdline.extend(
            ['--input-serialization', '{"CSV": {}, "CompressionType": "GZIP"}']
        )
        cmdline.extend(['--output-serialization', '{"CSV": {}}'])
        cmdline.extend([filename])

        expected_params = {
            'Bucket': 'mybucket',
            'Key': 'mykey',
            'Expression': 'SELECT * FROM S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {'CSV': {}, 'CompressionType': 'GZIP'},
            'OutputSerialization': {'CSV': {}},
            'RequestProgress': {'Enabled': True},
        }
        stdout = self.assert_params_for_cmd(cmdline, expected_params)[0]
        self.assertEqual(stdout, '')
        with open(filename) as f:
            contents = f.read()
            self.assertEqual(contents, ('a,b,c,d\n' 'e,f,g,h\n'))

    def test_errors_are_propagated(self):
        self.http_response.status_code = 400
        self.parsed_response = {
            'Error': {
                'Code': 'CastFailed',
                'Message': 'Attempt to convert from one data type to another',
            }
        }
        cmdline = self.prefix + [
            '--bucket',
            'mybucket',
            '--key',
            'mykey',
            '--expression',
            'SELECT * FROM S3Object',
            '--expression-type',
            'SQL',
            '--request-progress',
            'Enabled=True',
            '--input-serialization',
            '{"CSV": {}, "CompressionType": "GZIP"}',
            '--output-serialization',
            '{"CSV": {}}',
            os.path.join(self._tempdir, 'outfile'),
        ]
        expected_params = {
            'Bucket': 'mybucket',
            'Key': 'mykey',
            'Expression': 'SELECT * FROM S3Object',
            'ExpressionType': 'SQL',
            'InputSerialization': {'CSV': {}, 'CompressionType': 'GZIP'},
            'OutputSerialization': {'CSV': {}},
            'RequestProgress': {'Enabled': True},
        }
        self.assert_params_for_cmd(
            cmd=cmdline,
            params=expected_params,
            expected_rc=254,
            stderr_contains=(
                'An error occurred (CastFailed) when '
                'calling the SelectObjectContent operation'
            ),
        )


class TestHelpOutput(BaseAWSHelpOutputTest):
    def test_output(self):
        self.driver.main(['s3api', 'select-object-content', 'help'])
        # We don't want to be super picky because the wording may change
        # We just want to verify the Output section was customized.
        self.assert_contains(
            'Output\n======\n' 'This command generates no output'
        )
        self.assert_not_contains('[outfile')
        self.assert_contains('outfile')



import pytest

from awscli.customizations.ecs.expressgateway.managedresource import (
    ManagedResource,
)
from awscli.customizations.ecs.expressgateway.managedresourcegroup import (
    ManagedResourceGroup,
)


@pytest.fixture
def resource1():
    return ManagedResource("LoadBalancer", "lb-123", "ACTIVE", 1761230543.151)


@pytest.fixture
def resource2():
    return ManagedResource(
        "Certificate", "cert-456", "PROVISIONING", 1761230543.151
    )


class TestManagedResourceGroup:
    def test_is_terminal_all_terminal(self):
        terminal_resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        group = ManagedResourceGroup(resources=[terminal_resource])
        assert group.is_terminal()

    def test_is_terminal_mixed(self, resource1, resource2):
        group = ManagedResourceGroup(resources=[resource1, resource2])
        assert not group.is_terminal()

    def test_is_terminal_empty(self):
        group = ManagedResourceGroup()
        assert group.is_terminal()

    def test_get_status_string_with_header(self, resource1):
        group = ManagedResourceGroup(
            resource_type="IngressPaths",
            identifier="endpoint-1",
            resources=[resource1],
        )
        status_string = group.get_status_string("⠋")
        assert "IngressPaths" in status_string
        assert "endpoint-1" in status_string

    def test_compare_resource_sets_unique_resources(
        self, resource1, resource2
    ):
        # Test compare_resource_sets with completely different resources
        group1 = ManagedResourceGroup(resources=[resource1])  # LoadBalancer
        group2 = ManagedResourceGroup(resources=[resource2])  # Certificate

        diff1, diff2 = group1.compare_resource_sets(group2)

        # Each group should contain its unique resource
        assert len(diff1.resource_mapping) == 1
        assert len(diff2.resource_mapping) == 1
        assert "LoadBalancer/lb-123" in diff1.resource_mapping
        assert "Certificate/cert-456" in diff2.resource_mapping

    def test_compare_resource_sets_overlapping_resources(
        self, resource1, resource2
    ):
        # Test compare_resource_sets with same resource type but different identifiers
        resource3 = ManagedResource(
            "LoadBalancer", "lb-456", "FAILED", 1761230600.151
        )
        group1 = ManagedResourceGroup(
            resources=[resource1, resource2]
        )  # lb-123, cert-456
        group2 = ManagedResourceGroup(
            resources=[resource2, resource3]
        )  # cert-456, lb-456

        diff1, diff2 = group1.compare_resource_sets(group2)

        # group1 unique: lb-123, group2 unique: lb-456, common: cert-456 (should not appear in diff)
        assert len(diff1.resource_mapping) == 1
        assert len(diff2.resource_mapping) == 1
        assert "LoadBalancer/lb-123" in diff1.resource_mapping
        assert "LoadBalancer/lb-456" in diff2.resource_mapping
        # Common resource should not be in either diff
        assert "Certificate/cert-456" not in diff1.resource_mapping
        assert "Certificate/cert-456" not in diff2.resource_mapping

    def test_compare_resource_sets_identical_groups(
        self, resource1, resource2
    ):
        # Test compare_resource_sets with identical resource groups
        group1 = ManagedResourceGroup(resources=[resource1, resource2])
        group2 = ManagedResourceGroup(resources=[resource1, resource2])

        diff1, diff2 = group1.compare_resource_sets(group2)

        # No differences should be found
        assert len(diff1.resource_mapping) == 0
        assert len(diff2.resource_mapping) == 0

    def test_compare_resource_sets_empty_groups(self, resource1):
        # Test compare_resource_sets with empty groups
        group1 = ManagedResourceGroup(resources=[resource1])
        group2 = ManagedResourceGroup(resources=[])

        diff1, diff2 = group1.compare_resource_sets(group2)

        # group1 should contain its resource, group2 should be empty
        assert len(diff1.resource_mapping) == 1
        assert len(diff2.resource_mapping) == 0
        assert "LoadBalancer/lb-123" in diff1.resource_mapping

    def test_compare_resource_sets_excludes_matching_types_without_identifier(
        self,
    ):
        # Test that resources in other are excluded if self has same type without identifier
        resource_without_id = ManagedResource("LoadBalancer", None)
        resource_with_id = ManagedResource("LoadBalancer", "lb-456")

        group1 = ManagedResourceGroup(
            resources=[resource_without_id]
        )  # LoadBalancer/
        group2 = ManagedResourceGroup(
            resources=[resource_with_id]
        )  # LoadBalancer/lb-456

        diff1, diff2 = group1.compare_resource_sets(group2)

        # group1 should contain its resource without identifier
        assert len(diff1.resource_mapping) == 1
        assert "LoadBalancer/" in diff1.resource_mapping

        # group2 should be empty because LoadBalancer/lb-456 is excluded by LoadBalancer/
        assert len(diff2.resource_mapping) == 0

    def test_get_status_string_with_status(self):
        group = ManagedResourceGroup(
            resource_type="IngressPaths", identifier="test-id", status="ACTIVE"
        )
        result = group.get_status_string("⠋")
        assert "IngressPaths" in result
        assert "test-id" in result
        assert "✓" in result  # Green checkmark for ACTIVE
        assert "ACTIVE" in result

    def test_get_status_string_without_status(self):
        group = ManagedResourceGroup(
            resource_type="IngressPaths", identifier="test-id"
        )
        result = group.get_status_string("⠋")
        assert "IngressPaths" in result
        assert "test-id" in result
        assert "✓" not in result  # No symbol when no status
        assert "ACTIVE" not in result

    def test_get_status_string_status_without_identifier(self):
        group = ManagedResourceGroup(
            resource_type="IngressPaths", status="FAILED"
        )
        result = group.get_status_string("⠋")
        assert "IngressPaths" in result
        assert "X" in result  # Red X for FAILED
        assert "FAILED" in result

    def test_get_status_string_no_color(self, resource1):
        group = ManagedResourceGroup(
            resource_type="IngressPaths",
            identifier="test-id",
            status="ACTIVE",
            resources=[resource1],
        )
        result = group.get_status_string("⠋", use_color=False)
        assert "IngressPaths" in result
        assert "test-id" in result
        assert "✓" in result  # Checkmark should still be there
        assert "ACTIVE" in result
        # Should not contain ANSI color codes
        assert "\x1b[" not in result

    def test_get_status_string_with_color(self, resource1):
        group = ManagedResourceGroup(
            resource_type="IngressPaths",
            identifier="test-id",
            status="ACTIVE",
            resources=[resource1],
        )
        result = group.get_status_string("⠋", use_color=True)
        assert "IngressPaths" in result
        assert "test-id" in result
        assert "✓" in result  # Checkmark should be there
        assert "ACTIVE" in result
        # Should contain ANSI color codes
        assert "\x1b[" in result

    def test_combine_prioritizes_resources_with_identifier(self):
        resource_with_id = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        resource_without_id = ManagedResource(
            "LoadBalancer",
            None,
            "PROVISIONING",
            1761230600.151,  # newer timestamp
        )

        group1 = ManagedResourceGroup(
            resource_type="Service", resources=[resource_with_id]
        )
        group2 = ManagedResourceGroup(
            resource_type="Service", resources=[resource_without_id]
        )

        # Should prefer the one with identifier despite older timestamp
        result = group1.combine(group2)

        # Should only have the resource with identifier
        assert len(result.resource_mapping) == 1
        combined_resource = list(result.resource_mapping.values())[0]
        assert combined_resource.identifier == "lb-123"

    def test_get_stream_string_empty_group(self):
        """Test empty resource group returns empty string"""
        group = ManagedResourceGroup()
        result = group.get_stream_string("2025-12-15 10:00:00")
        assert result == ""

    def test_get_stream_string_single_resource(self):
        """Test resource group with single resource"""
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        group = ManagedResourceGroup(resources=[resource])
        result = group.get_stream_string("2025-12-15 10:00:00")
        assert "[2025-12-15 10:00:00]" in result
        assert "LoadBalancer" in result
        assert "lb-123" in result
        assert "ACTIVE" in result

    def test_get_stream_string_multiple_resources(self):
        """Test resource group with multiple resources"""
        resource1 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        resource2 = ManagedResource(
            "TargetGroup", "tg-456", "PROVISIONING", 1761230543.151
        )
        group = ManagedResourceGroup(resources=[resource1, resource2])
        result = group.get_stream_string("2025-12-15 10:00:00")

        # Should have both resources
        assert "LoadBalancer" in result
        assert "lb-123" in result
        assert "TargetGroup" in result
        assert "tg-456" in result

        # Should have newline between resources
        lines = result.split("\n")
        assert len(lines) > 1

    def test_get_stream_string_nested_groups(self):
        """Test resource group with nested groups"""
        resource1 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        resource2 = ManagedResource(
            "TargetGroup", "tg-456", "ACTIVE", 1761230543.151
        )
        nested_group = ManagedResourceGroup(resources=[resource2])

        group = ManagedResourceGroup(resources=[resource1, nested_group])
        result = group.get_stream_string("2025-12-15 10:00:00")

        # Should have flattened both resources
        assert "LoadBalancer" in result
        assert "lb-123" in result
        assert "TargetGroup" in result
        assert "tg-456" in result

    def test_get_stream_string_deeply_nested_groups(self):
        """Test resource group with multiple levels of nesting"""
        resource1 = ManagedResource(
            "Cluster", "cluster-1", "ACTIVE", 1761230543.151
        )
        resource2 = ManagedResource(
            "Service", "service-1", "ACTIVE", 1761230543.151
        )
        resource3 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )

        nested_level2 = ManagedResourceGroup(resources=[resource3])
        nested_level1 = ManagedResourceGroup(
            resources=[resource2, nested_level2]
        )
        group = ManagedResourceGroup(resources=[resource1, nested_level1])

        result = group.get_stream_string("2025-12-15 10:00:00")

        # Should have all resources flattened
        assert "Cluster" in result
        assert "Service" in result
        assert "LoadBalancer" in result

    def test_get_stream_string_with_resource_details(self):
        """Test that resource details are preserved in stream output"""
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "PROVISIONING",
            1761230543.151,
            "Waiting for DNS propagation",
            "DNS: example.elb.amazonaws.com",
        )
        group = ManagedResourceGroup(resources=[resource])
        result = group.get_stream_string("2025-12-15 10:00:00")

        assert "Reason: Waiting for DNS propagation" in result
        assert "Info: DNS: example.elb.amazonaws.com" in result
        assert "Last Updated At:" in result

    def test_get_stream_string_preserves_order(self):
        """Test that resource order is preserved in output"""
        resource1 = ManagedResource(
            "Cluster", "cluster-1", "ACTIVE", 1761230543.151
        )
        resource2 = ManagedResource(
            "Service", "service-1", "ACTIVE", 1761230543.151
        )
        resource3 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )

        group = ManagedResourceGroup(
            resources=[resource1, resource2, resource3]
        )
        result = group.get_stream_string("2025-12-15 10:00:00")

        # Find positions of each resource type in output
        cluster_pos = result.find("Cluster")
        service_pos = result.find("Service")
        lb_pos = result.find("LoadBalancer")

        # Order should be preserved
        assert cluster_pos < service_pos
        assert service_pos < lb_pos

    def test_get_stream_string_no_color(self):
        """Test stream string without color codes"""
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        group = ManagedResourceGroup(resources=[resource])
        result = group.get_stream_string(
            "2025-12-15 10:00:00", use_color=False
        )

        assert "LoadBalancer" in result
        assert "lb-123" in result
        # Should not contain ANSI color codes
        assert "\x1b[" not in result

    def test_get_stream_string_with_color(self):
        """Test stream string with color codes"""
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        group = ManagedResourceGroup(resources=[resource])
        result = group.get_stream_string("2025-12-15 10:00:00", use_color=True)

        assert "LoadBalancer" in result
        assert "lb-123" in result
        # Should contain ANSI color codes
        assert "\x1b[" in result

    def test_get_stream_string_mixed_resource_types(self):
        """Test group with various resource types and statuses"""
        resources = [
            ManagedResource("Cluster", "cluster-1", "ACTIVE", 1761230543.151),
            ManagedResource(
                "Service", "service-1", "UPDATING", 1761230543.151
            ),
            ManagedResource(
                "LoadBalancer",
                "lb-123",
                "PROVISIONING",
                1761230543.151,
                "Creating",
            ),
            ManagedResource(
                "TargetGroup", "tg-456", "FAILED", 1761230543.151, "Error"
            ),
        ]
        group = ManagedResourceGroup(resources=resources)
        result = group.get_stream_string("2025-12-15 10:00:00")

        # All resources should be present
        assert "Cluster" in result
        assert "Service" in result
        assert "LoadBalancer" in result
        assert "TargetGroup" in result

        # All statuses should be present
        assert "ACTIVE" in result
        assert "UPDATING" in result
        assert "PROVISIONING" in result
        assert "FAILED" in result

    def test_get_stream_string_with_group_metadata(self):
        """Test that group-level metadata doesn't affect stream output"""
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        group = ManagedResourceGroup(
            resource_type="ManagedResources",
            identifier="group-1",
            resources=[resource],
            status="ACTIVE",
            reason="All resources healthy",
        )
        result = group.get_stream_string("2025-12-15 10:00:00")

        # Should still show the actual resource, not group metadata
        assert "LoadBalancer" in result
        assert "lb-123" in result

        # Group metadata should not appear in stream output
        # (stream output is for individual resources only)
        assert "group-1" not in result

    def test_get_stream_string_empty_nested_group(self):
        """Test nested group that is empty is handled correctly"""
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        empty_nested = ManagedResourceGroup()

        group = ManagedResourceGroup(resources=[resource, empty_nested])
        result = group.get_stream_string("2025-12-15 10:00:00")

        # Should show the resource from non-empty group
        assert "LoadBalancer" in result
        # Empty nested group shouldn't add extra content
        lines = [line for line in result.split("\n") if line.strip()]
        # LoadBalancer resource produces multiple lines (timestamp line + optional detail lines)
        assert len(lines) > 0



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
from io import StringIO
from unittest.mock import patch

import pytest

from awscli.customizations.configure import (
    is_overly_permissive,
    warn_if_permissive,
)
from awscli.testutils import skip_if_windows


@pytest.mark.parametrize("mode", [0o700, 0o600, 0o400, 0o200, 0o000])
def test_acceptable_modes_are_not_overly_permissive(mode):
    assert is_overly_permissive(mode) is False


@pytest.mark.parametrize("mode", [0o644, 0o666, 0o777, 0o610, 0o601])
def test_overly_permissive_modes_are_detected(mode):
    assert is_overly_permissive(mode) is True


def _create_creds_file(tmp_path, mode):
    path = tmp_path / "credentials"
    path.write_text("[default]\naws_access_key_id=testing\n")
    os.chmod(path, mode)
    return str(path)


@skip_if_windows("Permissions test not valid on Windows.")
def test_prints_warning_for_permissive_file(tmp_path):
    path = _create_creds_file(tmp_path, 0o644)
    err = StringIO()
    warn_if_permissive(path, err_stream=err)
    output = err.getvalue()
    assert "aws: [WARNING]" in output
    assert path in output
    assert "accessible by other users" in output


@skip_if_windows("Permissions test not valid on Windows.")
def test_no_warning_for_0o600_file(tmp_path):
    path = _create_creds_file(tmp_path, 0o600)
    err = StringIO()
    warn_if_permissive(path, err_stream=err)
    assert err.getvalue() == ""


@skip_if_windows("Permissions test not valid on Windows.")
def test_no_warning_for_0o700_file(tmp_path):
    path = _create_creds_file(tmp_path, 0o700)
    err = StringIO()
    warn_if_permissive(path, err_stream=err)
    assert err.getvalue() == ""


@skip_if_windows("Permissions test not valid on Windows.")
def test_skips_when_file_does_not_exist(tmp_path):
    err = StringIO()
    warn_if_permissive(str(tmp_path / "nonexistent"), err_stream=err)
    assert err.getvalue() == ""


@patch("awscli.customizations.configure.is_windows", True)
def test_skips_on_windows(tmp_path):
    path = _create_creds_file(tmp_path, 0o777)
    err = StringIO()
    warn_if_permissive(path, err_stream=err)
    assert err.getvalue() == ""



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr.emrutils import which


class TestEMRutils:
    def test_which_with_existing_command(self):
        pythonPath = which('python') or which('python.exe')
        assert pythonPath is not None

    def test_which_with_non_existing_command(self):
        path = which('klajsflklj')
        assert path is None



#!/usr/bin/env python
# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Test help output for the AWS CLI.

The purpose of these docs is to test that the generated output looks how
we expect.

It's intended to be as end to end as possible, but instead of looking
at the man output, we look one step before at the generated rst output
(it's easier to verify).

"""

import os

import pytest

from awscli.alias import AliasLoader
from awscli.compat import StringIO
from awscli.testutils import (
    BaseAWSHelpOutputTest,
    CapturedRenderer,
    FileCreator,
    mock,
)
from tests import CLIRunner

COMMAND_ARGS_TEST_DATA = [
    {
        'command_args': ["ec2", "create-launch-template-version", "help"],
        'expected_url_suffix': "/reference/ec2/create-launch-template-version.html",
    },
    {
        'command_args': ["help"],
        'expected_url_suffix': "/index.html",
    },
    {
        'command_args': ["s3", "help"],
        'expected_url_suffix': "/reference/s3/index.html",
    },
]


def create_cases():
    for test_data in COMMAND_ARGS_TEST_DATA:
        yield pytest.param(test_data, id="-".join(test_data['command_args']))


def runner(config_file=None):
    runner = CLIRunner()

    # Add the PATH to the environment variables so that that posix help
    # renderers can find either the groff or mandoc executables required to
    # render the help pages for posix environments
    if "PATH" in os.environ:
        runner.env["PATH"] = os.environ["PATH"]

    if config_file is not None:
        runner.env['AWS_CONFIG_FILE'] = config_file

    return runner


@pytest.fixture
def runner_url():
    file_creator = FileCreator()
    return runner(
        file_creator.create_file(
            'config', '[default]\n' 'cli_help_output = url\n'
        )
    )


@pytest.fixture
def runner_browser():
    file_creator = FileCreator()
    return runner(
        file_creator.create_file(
            'config', '[default]\n' 'cli_help_output = browser\n'
        )
    )


class TestHelpOutput(BaseAWSHelpOutputTest):
    def test_output(self):
        self.driver.main(['help'])
        # Check for the reference label.
        self.assert_contains('.. _cli:aws:')
        self.assert_contains('***\naws\n***')
        self.assert_contains(
            'The AWS Command Line Interface is a unified tool '
            'to manage your AWS services.'
        )
        self.assert_contains('Use *aws help topics* to view')
        # Verify we see the docs for top level params, so pick
        # a few representative types of params.
        self.assert_contains('``--endpoint-url``')
        # Boolean type
        self.assert_contains('``--no-paginate``')
        # Arg with choices
        self.assert_contains('``--color``')
        self.assert_contains('*   on')
        self.assert_contains('*   off')
        self.assert_contains('*   auto')
        # Then we should see the services.
        self.assert_contains('* ec2')
        self.assert_contains('* s3api')
        self.assert_contains('* sts')
        # Make sure it its a related item
        self.assert_contains('========\nSee Also\n========')
        self.assert_contains('aws help topics')

    def test_service_help_output(self):
        self.driver.main(['ec2', 'help'])
        # Check for the reference label.
        self.assert_contains('.. _cli:aws ec2:')
        # We should see the section title for the service.
        self.assert_contains('***\nec2\n***')
        # With a description header.
        self.assert_contains('===========\nDescription\n===========')
        # And we should see the operations listed.
        self.assert_contains('* monitor-instances')
        self.assert_contains('* run-instances')
        self.assert_contains('* describe-instances')

    def test_operation_help_output(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        # Check for the reference label.
        self.assert_contains('.. _cli:aws ec2 run-instances:')
        # Should see the title with the operation name
        self.assert_contains('*************\nrun-instances\n*************')
        # Should contain part of the help text from the model.
        self.assert_contains('Launches the specified number of instances')
        self.assert_contains('``--count`` (string)')

    def test_custom_service_help_output(self):
        self.driver.main(['s3', 'help'])
        self.assert_contains('.. _cli:aws s3:')
        self.assert_contains('high-level S3 commands')
        self.assert_contains('* cp')

    def test_waiter_does_not_have_global_args(self):
        self.driver.main(['ec2', 'wait', 'help'])
        self.assert_not_contains('--debug')
        self.assert_not_contains('Global Options')

    def test_custom_operation_help_output(self):
        self.driver.main(['s3', 'ls', 'help'])
        self.assert_contains('.. _cli:aws s3 ls:')
        self.assert_contains('List S3 objects')
        self.assert_contains('--summarize')
        self.assert_contains('--debug')

    def test_topic_list_help_output(self):
        self.driver.main(['help', 'topics'])
        # Should contain the title
        self.assert_contains(
            '*******************\nAWS CLI Topic Guide\n*******************'
        )
        # Should contain the description
        self.assert_contains('This is the AWS CLI Topic Guide.')
        # Should contain the available topics section
        self.assert_contains('Available Topics')
        # Assert the general order of topic categories.
        self.assert_text_order(
            '-------\nGeneral\n-------',
            '--\nS3\n--',
            starting_from='Available Topics',
        )
        # Make sure that the topic elements elements show up as well.
        self.assert_contains('* return-codes: Describes')
        # Make sure the topic elements are underneath the categories as well
        # and they get added to each category they fall beneath
        self.assert_text_order(
            '-------\nGeneral\n-------',
            '* return-codes: Describes',
            '--\nS3\n--',
            starting_from='-------\nGeneral\n-------',
        )

    def test_topic_help_command(self):
        self.driver.main(['help', 'return-codes'])
        self.assert_contains(
            '********************\nAWS CLI Return Codes\n********************'
        )
        self.assert_contains('These are the following return codes')

    def test_arguments_with_example_json_syntax(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        self.assert_contains('``--iam-instance-profile``')
        self.assert_contains('JSON Syntax')
        self.assert_contains('"Arn": "string"')
        self.assert_contains('"Name": "string"')

    def test_arguments_with_example_shorthand_syntax(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        self.assert_contains('``--iam-instance-profile``')
        self.assert_contains('Shorthand Syntax')
        self.assert_contains('Arn=string,Name=string')

    def test_required_args_come_before_optional_args(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        # We're asserting that the args in the synopsis section appear
        # in this order.  They don't have to be in this exact order, but
        # each item in the list has to come before the previous arg.
        self.assert_text_order(
            '--image-id <value>',
            '[--key-name <value>]',
            '[--security-groups <value>]',
            starting_from='Synopsis',
        )

    def test_service_operation_order(self):
        self.driver.main(['ec2', 'help'])
        self.assert_text_order(
            'activate-license',
            'allocate-address',
            'assign-private-ip-addresses',
            starting_from='Available Commands',
        )

    def test_top_level_args_order(self):
        self.driver.main(['help'])
        self.assert_text_order(
            'autoscaling\n',
            'cloudformation\n',
            'elb\n',
            'swf\n',
            starting_from='Available Services',
        )

    def test_examples_in_operation_help(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        self.assert_contains('========\nExamples\n========')

    def test_add_help_for_dryrun(self):
        self.driver.main(['ec2', 'run-instances', 'help'])
        self.assert_contains('DryRunOperation')
        self.assert_contains('UnauthorizedOperation')

    def test_elb_help_output(self):
        self.driver.main(['elb', 'help'])
        # We should *not* have any invalid links like
        # .. _`:
        self.assert_not_contains('.. _`:')

    def test_shorthand_flattens_list_of_single_member_structures(self):
        self.driver.main(['elb', 'remove-tags', 'help'])
        self.assert_contains("--tags Key1 Key2 Key3")

    def test_deprecated_operations_not_documented(self):
        self.driver.main(['s3api', 'help'])
        self.assert_not_contains('get-bucket-lifecycle\n')
        self.assert_not_contains('put-bucket-lifecycle\n')
        self.assert_not_contains('get-bucket-notification\n')
        self.assert_not_contains('put-bucket-notification\n')


class TestRemoveDeprecatedCommands(BaseAWSHelpOutputTest):
    def assert_command_does_not_exist(self, service, command):
        # Basically try to get the help output for the removed
        # command verify that we get a SystemExit exception
        # and that we get something in stderr that says that
        # we made an invalid choice (because the operation is removed).
        stderr = StringIO()
        with mock.patch('sys.stderr', stderr):
            cr = self.driver.main([service, command, 'help'])
        self.assertEqual(cr, 252)
        # We should see an error message complaining about
        # an invalid choice because the operation has been removed.
        self.assertIn(
            'argument operation: Found invalid choice', stderr.getvalue()
        )

    def test_ses_deprecated_commands(self):
        self.driver.main(['ses', 'help'])
        self.assert_not_contains('list-verified-email-addresses')
        self.assert_not_contains('delete-verified-email-address')
        self.assert_not_contains('verify-email-address')

        self.assert_command_does_not_exist(
            'ses', 'list-verified-email-addresses'
        )
        self.assert_command_does_not_exist(
            'ses', 'delete-verified-email-address'
        )
        self.assert_command_does_not_exist('ses', 'verify-email-address')

    def test_ec2_import_export(self):
        self.driver.main(['ec2', 'help'])
        self.assert_not_contains('import-instance')
        self.assert_not_contains('import-volume')
        self.assert_command_does_not_exist('ec2', 'import-instance')
        self.assert_command_does_not_exist('ec2', 'import-volume')

    def test_boolean_param_documented(self):
        self.driver.main(
            ['autoscaling', 'terminate-instance-in-auto-scaling-group', 'help']
        )
        self.assert_contains(
            '``--should-decrement-desired-capacity`` | '
            '``--no-should-decrement-desired-capacity`` (boolean)'
        )

    def test_streaming_output_arg(self):
        self.driver.main(['s3api', 'get-object', 'help'])
        self.assert_not_contains('``--outfile``')
        self.assert_contains('``outfile`` (string)')

    def test_rds_add_arg_help_has_correct_command_name(self):
        self.driver.main(['rds', 'add-option-to-option-group', 'help'])
        self.assert_contains('add-option-to-option-group')

    def test_rds_remove_arg_help_has_correct_command_name(self):
        self.driver.main(['rds', 'remove-option-from-option-group', 'help'])
        self.assert_contains('remove-option-from-option-group')

    def test_modify_operation_not_in_help(self):
        self.driver.main(['rds', 'help'])
        # This was split into add/remove commands.  The modify
        # command should not be available.
        self.assert_not_contains('modify-option-group')


class TestPagingParamDocs(BaseAWSHelpOutputTest):
    def test_starting_token_injected(self):
        self.driver.main(['s3api', 'list-objects', 'help'])
        self.assert_contains('``--starting-token``')

    def test_max_items_injected(self):
        self.driver.main(['s3api', 'list-objects', 'help'])
        self.assert_contains('``--max-items``')

    def test_builtin_paging_params_removed(self):
        self.driver.main(['s3api', 'list-objects', 'help'])
        self.assert_not_contains('``--next-token``')
        self.assert_not_contains('``--max-keys``')

    def test_paging_documentation_added(self):
        self.driver.main(['s3api', 'list-objects', 'help'])
        self.assert_contains('``list-objects`` is a paginated operation')
        self.assert_contains('When using ``--output text`` and the')
        self.assert_contains('following query expressions: ')


class TestMergeBooleanGroupArgs(BaseAWSHelpOutputTest):
    def test_merge_bool_args(self):
        # Boolean args need to be group together so rather than
        # --foo foo docs
        # --no-foo foo docs again
        #
        # We instead have:
        # --foo | --no-foo foo docs
        self.driver.main(['ec2', 'run-instances', 'help'])
        self.assert_contains('``--dry-run`` | ``--no-dry-run``')

    def test_top_level_bools(self):
        # structure(scalar) of a single value of Value whose value is
        # a boolean is pulled into a top level arg.
        self.driver.main(['ec2', 'modify-instance-attribute', 'help'])
        self.assert_contains('``--ebs-optimized`` | ``--no-ebs-optimized``')

    def test_top_level_bool_has_no_example(self):
        # Normally a structure(bool) param would have an example
        # of {"Value": true|false}", but when we pull the arg up into
        # a top level bool, we should not generate an example.
        self.driver.main(['ec2', 'modify-instance-attribute', 'help'])
        self.assert_not_contains('"Value": true|false')


class TestStructureScalarHasNoExamples(BaseAWSHelpOutputTest):
    def test_no_examples_for_structure_single_scalar(self):
        self.driver.main(['ec2', 'modify-instance-attribute', 'help'])
        self.assert_not_contains('"Value": "string"')
        self.assert_not_contains('Value=string')

    def test_example_for_single_structure_not_named_value(self):
        # Verify that if a structure does match our special case
        # (single element named "Value"), then we still document
        # the example syntax.
        self.driver.main(['s3api', 'create-bucket', 'help'])
        self.assert_contains('LocationConstraint=string')
        # Also should see the JSON syntax in the help output.
        self.assert_contains('"LocationConstraint": ')


class TestJSONListScalarDocs(BaseAWSHelpOutputTest):
    def test_space_separated_list_docs(self):
        # A list of scalar type can be specified as JSON:
        #      JSON Syntax:
        #
        #       ["string", ...]
        # But at the same time you can always replace that with
        # a space separated list.  Therefore we want to document
        # the space separated list version and not the JSON list
        # version.
        self.driver.main(['ec2', 'terminate-instances', 'help'])
        self.assert_not_contains('["string", ...]')
        self.assert_contains('"string" "string"')


class TestParamRename(BaseAWSHelpOutputTest):
    def test_create_image_renames(self):
        # We're just cherry picking this particular operation to verify
        # that the rename arg customizations are working.
        self.driver.main(['ec2', 'create-image', 'help'])
        self.assert_not_contains('no-no-reboot')
        self.assert_contains('--reboot')


class TestCustomCommandDocsFromFile(BaseAWSHelpOutputTest):
    def test_description_from_rst_file(self):
        # The description for the configure command
        # is in _description.rst.  We're verifying that we
        # can read those contents properly.
        self.driver.main(['configure', 'help'])
        # These are a few options that are documented in the help output.
        self.assert_contains('metadata_service_timeout')
        self.assert_contains('metadata_service_num_attempts')
        self.assert_contains('aws_access_key_id')


class TestParametersCanBeHidden(BaseAWSHelpOutputTest):
    def mark_as_undocumented(self, argument_table, **kwargs):
        argument_table['starting-sequence-number']._UNDOCUMENTED = True

    def test_hidden_params_are_not_documented(self):
        # We're going to demonstrate hiding a parameter.
        # --device
        self.driver.session.register(
            'building-argument-table', self.mark_as_undocumented
        )
        self.driver.main(['kinesis', 'get-shard-iterator', 'help'])
        self.assert_not_contains('--starting-sequence-number')


class TestCanDocumentAsRequired(BaseAWSHelpOutputTest):
    def test_can_doc_as_required(self):
        # This param is already marked as required, but to be
        # explicit this is repeated here to make it more clear.
        def doc_as_required(argument_table, **kwargs):
            arg = argument_table['volume-arns']  # noqa: F841

        self.driver.session.register(
            'building-argument-table', doc_as_required
        )
        self.driver.main(
            ['storagegateway', 'describe-cached-iscsi-volumes', 'help']
        )
        self.assert_not_contains('[--volume-arns <value>]')


class TestEC2AuthorizeSecurityGroupNotRendered(BaseAWSHelpOutputTest):
    def test_deprecated_args_not_documented(self):
        self.driver.main(['ec2', 'authorize-security-group-ingress', 'help'])
        self.assert_not_contains('--ip-protocol')
        self.assert_not_contains('--from-port')
        self.assert_not_contains('--to-port')
        self.assert_not_contains('--source-security-group-name')
        self.assert_not_contains('--source-security-group-owner-id')


class TestKMSCreateGrant(BaseAWSHelpOutputTest):
    def test_proper_casing(self):
        self.driver.main(['kms', 'create-grant', 'help'])
        # Ensure that the proper casing is used for this command's docs.
        self.assert_not_contains('generate-data-key')
        self.assert_contains('GenerateDataKey')


class TestRoute53CreateHostedZone(BaseAWSHelpOutputTest):
    def test_proper_casing(self):
        self.driver.main(['route53', 'create-hosted-zone', 'help'])
        # Ensure that the proper casing is used for this command's docs.
        self.assert_contains(
            'do **not** include ``PrivateZone`` in this input structure'
        )


class TestIotData(BaseAWSHelpOutputTest):
    def test_service_help_command_has_note(self):
        self.driver.main(['iot-data', 'help'])
        # Ensure the note is in help page.
        self.assert_contains(
            'The default endpoints (intended for testing purposes only) can be found at '
            'https://docs.aws.amazon.com/general/latest/gr/iot-core.html#iot-core-data-plane-endpoints'
        )

    def test_operation_help_command_has_note(self):
        self.driver.main(['iot-data', 'get-thing-shadow', 'help'])
        # Ensure the note is in help page.
        self.assert_contains(
            'The default endpoints (intended for testing purposes only) can be found at '
            'https://docs.aws.amazon.com/general/latest/gr/iot-core.html#iot-core-data-plane-endpoints'
        )


class TestAliases(BaseAWSHelpOutputTest):
    def setUp(self):
        super().setUp()
        self.files = FileCreator()
        self.alias_file = self.files.create_file('alias', '[toplevel]\n')
        self.driver.alias_loader = AliasLoader(self.alias_file)

    def tearDown(self):
        super().tearDown()
        self.files.remove_all()

    def add_alias(self, alias_name, alias_value):
        with open(self.alias_file, 'a+') as f:
            f.write(f'{alias_name} = {alias_value}\n')

    def test_alias_not_in_main_help(self):
        self.add_alias('my-alias', 'ec2 describe-regions')
        self.driver.main(['help'])
        self.assert_not_contains('my-alias')


class TestStreamingOutputHelp(BaseAWSHelpOutputTest):
    def test_service_help_command_has_note(self):
        self.driver.main(['s3api', 'get-object', 'help'])
        self.assert_not_contains('outfile <value>')
        self.assert_contains('<outfile>')


class TestUrlOutputHelp:
    @pytest.mark.parametrize(
        "test_case",
        create_cases(),
    )
    @mock.patch('awscli.help.get_renderer')
    def test_docs_prints_url(self, mock_get_renderer, test_case, runner_url):
        renderer = CapturedRenderer()
        mock_get_renderer.return_value = renderer

        runner_url.run(test_case['command_args'])
        assert (
            "https://awscli.amazonaws.com/v2/documentation/api/"
            in renderer.rendered_contents
        )
        assert test_case['expected_url_suffix'] in renderer.rendered_contents


# Use this test class for "help" cases that require the default renderer
# (i.e. renderer from get_render()) instead of a mocked version.
class TestHelpOutputDefaultRenderer:
    def test_line_lengths_do_not_break_create_launch_template_version_cmd(
        self,
    ):
        runner = CLIRunner()
        # Add the PATH to the environment variables so that that posix help
        # renderers can find either the groff or mandoc executables required to
        # render the help pages for posix environments
        if "PATH" in os.environ:
            runner.env["PATH"] = os.environ["PATH"]

        result = runner.run(["ec2", "create-launch-template-version", "help"])
        assert 'exceeds the line-length-limit' not in result.stderr


@pytest.mark.skip("Cross-test interaction with CLIRunner mocking os.environ")
class TestHelpOutputBrowserRenderer:
    @pytest.mark.parametrize("test_case", create_cases())
    @mock.patch("awscli.help.webbrowser.open_new_tab")
    def test_docs_opens_browser(
        self, mock_open_new_tab, test_case, runner_browser
    ):
        runner_result = runner_browser.run(test_case['command_args'])
        assert (
            "Opening help file in the default browser." in runner_result.stdout
        )
        mock_open_new_tab.assert_called_once()



import os
import shutil
import tempfile

from awscli.testutils import unittest
from awscli.utils import write_exception


class TestWriteException(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.outfile = os.path.join(self.tempdir, 'stdout')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_write_exception(self):
        error_message = "Some error message."
        ex = Exception(error_message)
        with open(self.outfile, 'w+', encoding='utf-8') as outfile:
            write_exception(ex, outfile)
            outfile.seek(0)

            expected_output = "\n%s\n" % error_message
            self.assertEqual(outfile.read(), expected_output)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document

from awscli.autocomplete.completer import CompletionResult
from awscli.autoprompt.logger import PromptToolkitHandler
from awscli.autoprompt.prompttoolkit import (
    PromptToolkitCompleter,
    loggers_handler_switcher,
)
from awscli.testutils import mock, unittest


class TestPromptToolkitCompleter(unittest.TestCase):
    def setUp(self):
        self.completion_source = mock.Mock()
        self.completion_source.autocomplete = mock.Mock()

    def assert_completions_match_expected(
        self, actual_completions, expected_completion_objects
    ):
        expected_completions = (
            completion for completion in expected_completion_objects
        )
        for actual, expected in zip(actual_completions, expected_completions):
            self.assertEqual(actual.text, expected.text)
            self.assertEqual(actual.start_position, expected.start_position)
            self.assertEqual(actual.display, expected.display)

    def test_get_completions(self):
        expected_completion_objects = [
            Completion('create-image', 0, 'create-image', ''),
            Completion('describe-instances', 0, 'describe-instances', ''),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult('create-image'),
            CompletionResult('describe-instances'),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )

    def test_get_completions_has_no_completions(self):
        self.completion_source.autocomplete.return_value = []
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assertRaises(StopIteration, lambda: next(actual_completions))

    def test_get_completions_sorted_by_required(self):
        expected_completion_objects = [
            Completion('--name', 0, '--name (required)', ''),
            Completion('--instance-id', 0, '--instance-id (required)', ''),
            Completion('--debug', 0, '--debug', ''),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult('--debug', required=False),
            CompletionResult('--name', required=True),
            CompletionResult('--instance-id', required=True),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )

    def test_get_completions_with_auto_prompt_overrides_filtered_out(self):
        expected_completion_objects = [
            Completion('--name', 0, '--name (required)', ''),
            Completion('--instance-id', 0, '--instance-id (required)', ''),
            Completion('--debug', 0, '--debug', ''),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult('--debug', required=False),
            CompletionResult('--name', required=True),
            CompletionResult('--instance-id', required=True),
            CompletionResult('--cli-auto-prompt', required=False),
            CompletionResult('--no-cli-auto-prompt', required=False),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )

    def test_get_completions_with_duplicates_removed(self):
        expected_completion_objects = [
            Completion('--name', 0, '--name (required)', ''),
            Completion('--instance-id', 0, '--instance-id (required)', ''),
            Completion('--debug', 0, '--debug', ''),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult('--debug', required=False),
            CompletionResult('--debug', required=False),
            CompletionResult('--name', required=True),
            CompletionResult('--name', required=True),
            CompletionResult('--instance-id', required=True),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )

    def test_get_completions_with_overrides_and_duplicates_removed(self):
        expected_completion_objects = [
            Completion('--name', 0, '--name (required)', ''),
            Completion('--instance-id', 0, '--instance-id (required)', ''),
            Completion('--debug', 0, '--debug', ''),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult('--debug', required=False),
            CompletionResult('--debug', required=False),
            CompletionResult('--name', required=True),
            CompletionResult('--name', required=True),
            CompletionResult('--instance-id', required=True),
            CompletionResult('--cli-auto-prompt', required=False),
            CompletionResult('--no-cli-auto-prompt', required=False),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )

    def test_get_completions_with_display_meta(self):
        expected_completion_objects = [
            Completion(
                '--name',
                0,
                '--name (required)',
                '[string] A name for the new image.',
            ),
            Completion(
                '--instance-id',
                0,
                '--instance-id (required)',
                '[string] The ID of the instance.',
            ),
            Completion(
                '--debug', 0, '--debug', '[boolean] Turn on debug logging.'
            ),
        ]
        self.completion_source.autocomplete.return_value = [
            CompletionResult(
                '--debug',
                required=False,
                cli_type_name='string',
                help_text='The ID of the instance.',
            ),
            CompletionResult(
                '--name',
                required=True,
                cli_type_name='string',
                help_text='A name for the new image.',
            ),
            CompletionResult(
                '--instance-id',
                required=True,
                cli_type_name='boolean',
                help_text='Turn on debug logging.',
            ),
        ]
        self.completer = PromptToolkitCompleter(self.completion_source)
        actual_completions = self.completer.get_completions(
            Document(), CompleteEvent()
        )
        self.assert_completions_match_expected(
            actual_completions, expected_completion_objects
        )


class TestLoggersHandlerSwitcher(unittest.TestCase):
    def test_can_switch_and_restore_handlers(self):
        log = logging.getLogger('test_prompt_logger')
        log.handlers = []
        log.addHandler(logging.StreamHandler())
        log.addHandler(logging.NullHandler())
        with loggers_handler_switcher():
            handlers = log.handlers
            self.assertEqual(len(handlers), 1)
            self.assertIsInstance(handlers[0], PromptToolkitHandler)
        handlers = log.handlers
        self.assertEqual(len(handlers), 2)
        self.assertIsInstance(handlers[0], logging.StreamHandler)
        self.assertIsInstance(handlers[1], logging.NullHandler)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr.createdefaultroles import get_service_principal
from awscli.customizations.emr.exceptions import ResolveServicePrincipalError
from awscli.testutils import unittest


class TestEmrConfig(unittest.TestCase):
    ec2_service = "ec2"
    ec2_service_principal = "ec2.amazonaws.com"
    emr_service = "elasticmapreduce"
    endpoint1 = "https://ap-southeast-1.elasticmapreduce.abc/"
    endpoint2 = "https://elasticmapreduce.abcd.def.ghi"
    endpoint3 = "https://garbage.nothing.com"
    expected_result1 = "elasticmapreduce.abc"
    expected_result2 = "elasticmapreduce.def.ghi"

    def test_get_emr_service_principal(self):
        msg = (
            "Generated Service Principal does not match the expected"
            + "Service Principal"
        )

        result1 = get_service_principal(self.emr_service, self.endpoint1)
        self.assertEqual(result1, self.expected_result1, msg)

        result2 = get_service_principal(self.emr_service, self.endpoint2)
        self.assertEqual(result2, self.expected_result2, msg)

        self.assertRaises(
            ResolveServicePrincipalError,
            get_service_principal,
            self.emr_service,
            self.endpoint3,
        )

    def test_get_ec2_service_principal(self):
        self.assertEqual(
            get_service_principal(self.ec2_service, self.endpoint1),
            self.ec2_service_principal,
        )
        self.assertEqual(
            get_service_principal(self.ec2_service, self.endpoint2),
            self.ec2_service_principal,
        )
        self.assertEqual(
            get_service_principal(self.ec2_service, self.endpoint3),
            self.ec2_service_principal,
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestTerminateInstanceInAutoscalingGroup(BaseAWSCommandParamsTest):
    PREFIX = 'autoscaling terminate-instance-in-auto-scaling-group'

    def test_true(self):
        cmdline = self.PREFIX
        cmdline += ' --instance-id i-12345678'
        cmdline += ' --should-decrement-desired-capacity'
        params = {
            'InstanceId': 'i-12345678',
            'ShouldDecrementDesiredCapacity': True,
        }
        self.assert_params_for_cmd(cmdline, params)

    def test_false(self):
        cmdline = self.PREFIX
        cmdline += ' --instance-id i-12345678'
        cmdline += ' --no-should-decrement-desired-capacity'
        params = {
            'InstanceId': 'i-12345678',
            'ShouldDecrementDesiredCapacity': False,
        }
        self.assert_params_for_cmd(cmdline, params)

    def test_last_arg_wins(self):
        cmdline = self.PREFIX
        cmdline += ' --instance-id i-12345678'
        cmdline += ' --should-decrement-desired-capacity'
        cmdline += ' --no-should-decrement-desired-capacity'
        # Since the --no-should-decrement-desired-capacity was
        # was added last, it wins.
        params = {
            'InstanceId': 'i-12345678',
            'ShouldDecrementDesiredCapacity': False,
        }
        self.assert_params_for_cmd(cmdline, params)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.emr.createdefaultroles import assume_role_policy
from awscli.testutils import unittest


class TestDefaultRoles(unittest.TestCase):
    service_principal = "ec2.amazonaws.com"
    expected_result = {
        "Version": "2008-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"Service": "ec2.amazonaws.com"},
                "Action": "sts:AssumeRole",
            }
        ],
    }

    def test_assume_role_policy(self):
        result = assume_role_policy(self.service_principal)
        self.assertEqual(result, self.expected_result)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.ecs.deploy import (
    TIMEOUT_BUFFER_MIN,
    CodeDeployValidator,
)
from awscli.customizations.ecs.exceptions import (
    InvalidPlatformError,
    InvalidProperyError,
)
from awscli.testutils import unittest


class TestCodeDeployValidator(unittest.TestCase):
    TEST_RESOURCES = {
        'service': 'test-service',
        'service_arn': 'arn:aws:ecs:::service/test-service',
        'cluster': 'test-cluster',
        'cluster_arn': 'arn:aws:ecs:::cluster/test-cluster',
        'app_name': 'test-application',
        'deployment_group_name': 'test-deployment-group',
    }

    TEST_APP_DETAILS = {
        'application': {
            'applicationId': '876uyh6-45tdfg',
            'applicationName': 'test-application',
            'computePlatform': 'ECS',
        }
    }

    TEST_DEPLOYMENT_GROUP_DETAILS = {
        'deploymentGroupInfo': {
            'applicationName': 'test-application',
            'deploymentGroupName': 'test-deployment-group',
            'computePlatform': 'ECS',
            'blueGreenDeploymentConfiguration': {
                'deploymentReadyOption': {'waitTimeInMinutes': 5},
                'terminateBlueInstancesOnDeploymentSuccess': {
                    'terminationWaitTimeInMinutes': 10
                },
            },
            'ecsServices': [
                {'serviceName': 'test-service', 'clusterName': 'test-cluster'}
            ],
        }
    }

    def setUp(self):
        self.validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        self.validator.app_details = self.TEST_APP_DETAILS
        self.validator.deployment_group_details = (
            self.TEST_DEPLOYMENT_GROUP_DETAILS
        )

    def test_get_deployment_wait_time(self):
        expected_wait = 5 + 10 + TIMEOUT_BUFFER_MIN
        actual_wait = self.validator.get_deployment_wait_time()
        self.assertEqual(expected_wait, actual_wait)

    def test_get_deployment_wait_time_no_dgp(self):
        empty_validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        actual_wait = empty_validator.get_deployment_wait_time()
        self.assertEqual(None, actual_wait)

    def test_validations(self):
        self.validator.validate_application()
        self.validator.validate_deployment_group()
        self.validator.validate_all()

    def test_validate_application_error_compute_platform(self):
        invalid_app = {
            'application': {
                'applicationName': 'test-application',
                'computePlatform': 'Server',
            }
        }

        bad_validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        bad_validator.app_details = invalid_app

        with self.assertRaises(InvalidPlatformError):
            bad_validator.validate_application()

    def test_validate_deployment_group_error_compute_platform(self):
        invalid_dgp = {'deploymentGroupInfo': {'computePlatform': 'Lambda'}}
        bad_validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        bad_validator.deployment_group_details = invalid_dgp

        with self.assertRaises(InvalidPlatformError):
            bad_validator.validate_deployment_group()

    def test_validate_deployment_group_error_service(self):
        invalid_dgp = {
            'deploymentGroupInfo': {
                'computePlatform': 'ECS',
                'ecsServices': [
                    {
                        'serviceName': 'the-wrong-test-service',
                        'clusterName': 'test-cluster',
                    }
                ],
            }
        }
        bad_validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        bad_validator.deployment_group_details = invalid_dgp

        with self.assertRaises(InvalidProperyError):
            bad_validator.validate_deployment_group()

    def test_validate_deployment_group_error_cluster(self):
        invalid_dgp = {
            'deploymentGroupInfo': {
                'computePlatform': 'ECS',
                'ecsServices': [
                    {
                        'serviceName': 'test-service',
                        'clusterName': 'the-wrong-test-cluster',
                    }
                ],
            }
        }
        bad_validator = CodeDeployValidator(None, self.TEST_RESOURCES)
        bad_validator.deployment_group_details = invalid_dgp

        with self.assertRaises(InvalidProperyError):
            bad_validator.validate_deployment_group()



import asyncio
from unittest.mock import Mock, patch

import pytest

from awscli.customizations.ecs.prompt_toolkit_display import Display


class TestPromptToolkitDisplay:
    @pytest.fixture
    def display(self, ptk_app_session):
        yield Display()

    def test_init(self, display):
        """Test Display initialization."""
        assert display.control is not None
        assert display.window is not None
        assert display.status_control is not None
        assert display.app is not None
        assert display.content_lines == 0
        assert display.raw_text == ""

    def test_display_updates_content(self, display):
        """Test display method updates content and line count."""
        test_text = "Line 1\nLine 2\nLine 3"
        test_status = "Status message"

        display.display(test_text, test_status)

        assert display.content_lines == 3
        assert display.raw_text == test_text
        assert display.status_control.text == test_status

    def test_display_without_status(self, display):
        """Test display method with only text."""
        test_text = "Single line"

        display.display(test_text)

        assert display.content_lines == 1
        assert display.raw_text == test_text

    @patch('awscli.customizations.ecs.prompt_toolkit_display.ANSI')
    def test_display_uses_ansi_formatting(self, mock_ansi, display):
        """Test display method uses ANSI for color formatting."""
        test_text = "Colored text"

        display.display(test_text)

        mock_ansi.assert_called_once_with(test_text)
        assert display.raw_text == test_text

    def test_scroll_bounds_calculation(self, display):
        """Test that content_lines is calculated correctly for scroll bounds."""
        multiline_text = "Line 1\nLine 2\nLine 3\nLine 4\nLine 5"

        display.display(multiline_text)

        assert display.content_lines == 5

    @patch('awscli.customizations.ecs.prompt_toolkit_display.Application')
    def test_run_calls_app_run_async(self, mock_app_class, display):
        """Test run method calls app.run_async()."""
        mock_app = Mock()
        mock_app_class.return_value = mock_app

        # Make run_async return a coroutine
        async def mock_run_async():
            return None

        mock_app.run_async = mock_run_async

        # Replace the display's app with our mock
        display.app = mock_app

        # Run the async method
        asyncio.run(display.run())

        # Verify the mock was set up correctly
        assert mock_app.run_async == mock_run_async

    def test_scrollable_pane_with_scrollbar(self, display):
        """Test that ScrollablePane is configured with scrollbar for better UX."""
        # The window should be wrapped in a ScrollablePane with scrollbar
        assert hasattr(display.window, 'show_scrollbar')
        # The inner window should have wrap_lines enabled
        inner_window = display.window.content
        assert hasattr(inner_window, 'wrap_lines')
        assert inner_window.wrap_lines is not None

    def test_long_line_wrapping(self, display):
        """Test that very long lines are handled properly with wrapping."""
        long_line = "This is a very long line that should wrap properly " * 10
        test_text = f"Short line\n{long_line}\nAnother short line"

        display.display(test_text)

        assert display.content_lines == 3
        # Verify the content is set correctly
        assert display.control.text is not None

    def test_display_handles_ansi_content_without_errors(self, display):
        """Test that display handles ANSI-colored content without errors."""
        # Content with various ANSI codes
        ansi_text = "\x1b[32mGreen\x1b[0m\n\x1b[31mRed\x1b[32mGreen\x1b[0m\n\x1b[1m\x1b[41mBold on Red\x1b[0m"

        # Should not raise any exceptions
        display.display(ansi_text)

        # Content should be stored and processed
        assert display.raw_text == ansi_text
        assert display.content_lines == 3

    def test_scroll_validation_with_ansi_content(self, display):
        """Test scroll position validation works with ANSI-colored content."""
        # Mock the app output for terminal size
        mock_output = Mock()
        mock_output.get_size.return_value = Mock(rows=20, columns=80)
        display.app.output = mock_output

        # Set content with ANSI codes
        ansi_text = "\x1b[32mShort line\x1b[0m\n\x1b[31mAnother line\x1b[0m"

        # Should not raise an exception and content should be stored
        display.display(ansi_text)
        assert display.raw_text == ansi_text

    def test_scroll_validation_handles_missing_output(self, display):
        """Test scroll validation gracefully handles missing app output."""
        # Set up display without proper app output
        display.app.output = None

        # Should not raise an exception
        display.display("Test content")
        assert display.raw_text == "Test content"

    def test_scroll_validation_handles_exceptions(self, display):
        """Test scroll validation clamps scroll position on terminal size exceptions."""
        mock_output = Mock()
        mock_output.get_size.side_effect = OSError("Terminal unavailable")
        display.app.output = mock_output

        display.window.vertical_scroll = 5

        display.display("Test content")
        assert display.window.vertical_scroll == 5

    def test_scroll_validation_with_long_lines(self, display):
        """Test scroll validation with lines that wrap."""
        # Mock the app output for terminal size
        mock_output = Mock()
        mock_output.get_size.return_value = Mock(rows=20, columns=80)
        display.app.output = mock_output

        # Create content with very long lines that will wrap
        long_line = "A" * 100  # Longer than terminal width
        content = f"Short\n{long_line}\nShort again"

        # Should handle wrapping without errors
        display.display(content)
        assert display.raw_text == content



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestReplaceNetworkACLEntry(BaseAWSCommandParamsTest):
    prefix = 'ec2 replace-network-acl-entry'

    def test_tcp(self):
        cmdline = self.prefix
        cmdline += ' --network-acl-id acl-12345678'
        cmdline += ' --rule-number 100'
        cmdline += ' --protocol tcp'
        cmdline += ' --rule-action allow'
        cmdline += ' --ingress'
        cmdline += ' --port-range From=22,To=22'
        cmdline += ' --cidr-block 0.0.0.0/0'
        result = {
            'NetworkAclId': 'acl-12345678',
            'RuleNumber': 100,
            'Protocol': '6',
            'RuleAction': 'allow',
            'Egress': False,
            'CidrBlock': '0.0.0.0/0',
            'PortRange': {'From': 22, 'To': 22},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_udp(self):
        cmdline = self.prefix
        cmdline += ' --network-acl-id acl-12345678'
        cmdline += ' --rule-number 100'
        cmdline += ' --protocol udp'
        cmdline += ' --rule-action allow'
        cmdline += ' --ingress'
        cmdline += ' --port-range From=22,To=22'
        cmdline += ' --cidr-block 0.0.0.0/0'
        result = {
            'NetworkAclId': 'acl-12345678',
            'RuleNumber': 100,
            'Protocol': '17',
            'RuleAction': 'allow',
            'Egress': False,
            'CidrBlock': '0.0.0.0/0',
            'PortRange': {'From': 22, 'To': 22},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_icmp(self):
        cmdline = self.prefix
        cmdline += ' --network-acl-id acl-12345678'
        cmdline += ' --rule-number 100'
        cmdline += ' --protocol icmp'
        cmdline += ' --rule-action allow'
        cmdline += ' --ingress'
        cmdline += ' --port-range From=22,To=22'
        cmdline += ' --cidr-block 0.0.0.0/0'
        result = {
            'NetworkAclId': 'acl-12345678',
            'RuleNumber': 100,
            'Protocol': '1',
            'RuleAction': 'allow',
            'Egress': False,
            'CidrBlock': '0.0.0.0/0',
            'PortRange': {'From': 22, 'To': 22},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_all(self):
        cmdline = self.prefix
        cmdline += ' --network-acl-id acl-12345678'
        cmdline += ' --rule-number 100'
        cmdline += ' --protocol all'
        cmdline += ' --rule-action allow'
        cmdline += ' --ingress'
        cmdline += ' --port-range From=22,To=22'
        cmdline += ' --cidr-block 0.0.0.0/0'
        result = {
            'NetworkAclId': 'acl-12345678',
            'RuleNumber': 100,
            'Protocol': '-1',
            'RuleAction': 'allow',
            'Egress': False,
            'CidrBlock': '0.0.0.0/0',
            'PortRange': {'From': 22, 'To': 22},
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_number(self):
        cmdline = self.prefix
        cmdline += ' --network-acl-id acl-12345678'
        cmdline += ' --rule-number 100'
        cmdline += ' --protocol 99'
        cmdline += ' --rule-action allow'
        cmdline += ' --ingress'
        cmdline += ' --port-range From=22,To=22'
        cmdline += ' --cidr-block 0.0.0.0/0'
        result = {
            'NetworkAclId': 'acl-12345678',
            'RuleNumber': 100,
            'Protocol': '99',
            'RuleAction': 'allow',
            'Egress': False,
            'CidrBlock': '0.0.0.0/0',
            'PortRange': {'From': 22, 'To': 22},
        }
        self.assert_params_for_cmd(cmdline, result)



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
import os
import re

import awscli.clidriver
from awscli.testutils import BaseAWSCommandParamsTest, FileCreator

# file is gone in python3, so instead IOBase must be used.
# Given this test module is the only place that cares about
# this type check, we do the check directly in this test module.
try:
    file_type = file
except NameError:
    import io

    file_type = io.IOBase


class TestPutObject(BaseAWSCommandParamsTest):
    maxDiff = None
    prefix = 's3api put-object'

    def setUp(self):
        super(TestPutObject, self).setUp()
        self.file_path = os.path.join(
            os.path.dirname(__file__), 'test_put_object_data'
        )
        self.files = FileCreator()

    def tearDown(self):
        super(TestPutObject, self).tearDown()
        self.files.remove_all()

    def test_simple(self):
        cmdline = self.prefix
        cmdline += ' --bucket mybucket'
        cmdline += ' --key mykey'
        cmdline += ' --body %s' % self.file_path
        result = {
            'uri_params': {'Bucket': 'mybucket', 'Key': 'mykey'},
            'headers': {'Expect': '100-continue'},
        }
        expected = {'Bucket': 'mybucket', 'Key': 'mykey'}
        self.assert_params_for_cmd(cmdline, expected, ignore_params=['Body'])
        self.assertEqual(self.last_kwargs['Body'].name, self.file_path)

    def test_headers(self):
        cmdline = self.prefix
        cmdline += ' --bucket mybucket'
        cmdline += ' --key mykey'
        cmdline += ' --body %s' % self.file_path
        cmdline += ' --acl public-read'
        cmdline += ' --content-encoding x-gzip'
        cmdline += ' --content-type text/plain'
        expected = {
            'ACL': 'public-read',
            'Bucket': 'mybucket',
            'ContentEncoding': 'x-gzip',
            'ContentType': 'text/plain',
            'Key': 'mykey',
        }
        self.assert_params_for_cmd(cmdline, expected, ignore_params=['Body'])
        self.assertEqual(self.last_kwargs['Body'].name, self.file_path)

    def test_website_redirect(self):
        cmdline = self.prefix
        cmdline += ' --bucket mybucket'
        cmdline += ' --key mykey'
        cmdline += ' --acl public-read'
        cmdline += ' --website-redirect-location http://www.example.com/'
        expected = {
            'ACL': 'public-read',
            'Bucket': 'mybucket',
            'Key': 'mykey',
            'WebsiteRedirectLocation': 'http://www.example.com/',
        }
        self.assert_params_for_cmd(cmdline, expected)

    def test_sse_key_with_binary_file(self):
        # Create contents that do not get mapped to ascii
        contents = b'\xc2'
        filename = self.files.create_file('key', contents, mode='wb')
        cmdline = self.prefix
        cmdline += ' --bucket mybucket'
        cmdline += ' --key mykey'
        cmdline += ' --sse-customer-algorithm AES256'
        cmdline += ' --sse-customer-key fileb://%s' % filename
        expected = {
            'Bucket': 'mybucket',
            'Key': 'mykey',
            'SSECustomerAlgorithm': 'AES256',
            'SSECustomerKey': 'wg==',  # Note the key gets base64 encoded.
            'SSECustomerKeyMD5': 'ZGXa0dMXUr4/MoPo9w/u9w==',
        }
        self.assert_params_for_cmd(cmdline, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import contextlib
import io
import logging
import os
import platform
import re
import sys
from collections import namedtuple

import awscrt.io
import botocore.model
import pytest
from botocore import xform_name
from botocore.awsrequest import AWSResponse
from botocore.compat import OrderedDict
from botocore.configprovider import (
    ConfigChainFactory,
    ConfigValueStore,
    create_botocore_default_config_mapping,
)
from botocore.exceptions import NoCredentialsError
from botocore.hooks import HierarchicalEmitter

import awscli
from awscli import formatter
from awscli.argparser import HELP_BLURB
from awscli.clidriver import (
    CLICommand,
    CLIDriver,
    CustomArgument,
    ServiceCommand,
    ServiceOperation,
    construct_cli_error_handlers_chain,
    create_clidriver,
    resolve_auto_prompt_mode,
    validate_auto_prompt_args_are_mutually_exclusive,
)
from awscli.compat import StringIO
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError
from awscli.paramfile import URIArgumentHandler
from awscli.testutils import BaseAWSCommandParamsTest, mock, unittest

GET_DATA = {
    'cli': {
        'description': 'description',
        'synopsis': 'usage: foo',
        'options': {
            "debug": {"action": "store_true", "help": "Turn on debug logging"},
            "output": {
                "choices": ["json", "text", "table"],
                "metavar": "output_format",
            },
            "query": {"help": ""},
            "profile": {"help": "", "metavar": "profile_name"},
            "region": {"metavar": "region_name"},
            "endpoint-url": {"help": "", "metavar": "endpoint_url"},
            "no-verify-ssl": {
                "action": "store_false",
                "dest": "verify_ssl",
                "help": "",
            },
            "no-paginate": {
                "action": "store_false",
                "help": "",
                "dest": "paginate",
            },
            "page-size": {
                "type": "int",
                "help": "",
            },
            "read-timeout": {"type": "int", "help": ""},
            "connect-timeout": {"type": "int", "help": ""},
        },
    },
}

GET_VARIABLE = {
    'provider': 'aws',
    'output': 'json',
    'api_versions': {},
    'pager': 'less',
}


MINI_SERVICE = {
    "metadata": {
        "apiVersion": "2006-03-01",
        "endpointPrefix": "s3",
        "globalEndpoint": "s3.amazonaws.com",
        "signatureVersion": "s3",
        "protocol": "rest-xml",
    },
    "operations": {
        "ListObjects": {
            "name": "ListObjects",
            "http": {"method": "GET", "requestUri": "/{Bucket}"},
            "input": {"shape": "ListObjectsRequest"},
            "output": {"shape": "ListObjectsOutput"},
        },
        "IdempotentOperation": {
            "name": "IdempotentOperation",
            "http": {"method": "GET", "requestUri": "/{Bucket}"},
            "input": {"shape": "IdempotentOperationRequest"},
            "output": {"shape": "ListObjectsOutput"},
        },
    },
    "shapes": {
        "ListObjectsOutput": {
            "type": "structure",
            "members": {
                "IsTruncated": {"shape": "IsTruncated", "documentation": ""},
                "NextMarker": {
                    "shape": "NextMarker",
                },
                "Contents": {"shape": "Contents"},
            },
        },
        "IdempotentOperationRequest": {
            "type": "structure",
            "required": "token",
            "members": {
                "token": {
                    "shape": "Token",
                    "idempotencyToken": True,
                },
            },
        },
        "ListObjectsRequest": {
            "type": "structure",
            "required": ["Bucket"],
            "members": OrderedDict(
                [
                    (
                        "Bucket",
                        {
                            "shape": "BucketName",
                            "location": "uri",
                            "locationName": "Bucket",
                        },
                    ),
                    (
                        "Marker",
                        {
                            "shape": "Marker",
                            "location": "querystring",
                            "locationName": "marker",
                        },
                    ),
                    (
                        "MaxKeys",
                        {
                            "shape": "MaxKeys",
                            "location": "querystring",
                            "locationName": "max-keys",
                        },
                    ),
                ]
            ),
        },
        "BucketName": {"type": "string"},
        "MaxKeys": {"type": "integer"},
        "Marker": {"type": "string"},
        "IsTruncated": {"type": "boolean"},
        "NextMarker": {"type": "string"},
        "Contents": {"type": "string"},
        "Token": {"type": "string"},
    },
}


def _generate_auto_prompt_resolve_cases():
    # Each case is a 5-namedtuple with the following meaning:
    # "args" is a list of arguments that command got as input from
    #   command line
    # "config_variable" is the result from get_config_variable
    #   This takes a value of either 'on' , 'off' or 'on-partial'
    # "expected_result" is a boolean indicating whether auto-prompt
    #   should be used or not.
    #
    # Note: This set of tests assumes that only one of --no-cli-auto-prompt
    # or --cli-auto-prompt overrides can be specified.
    # TestCLIAutoPrompt.test_throw_error_if_both_args_specified tests
    # that these command line overrides are mutually exclusive.
    Case = namedtuple(
        'Case',
        [
            'args',
            'config_variable',
            'expected_result',
        ],
    )
    return [
        Case([], 'off', 'off'),
        Case([], 'on', 'on'),
        Case(['--cli-auto-prompt'], 'off', 'on'),
        Case(['--cli-auto-prompt'], 'on', 'on'),
        Case(['--no-cli-auto-prompt'], 'off', 'off'),
        Case(['--no-cli-auto-prompt'], 'on', 'off'),
        Case([], 'on', 'on'),
        Case([], 'on-partial', 'on-partial'),
        Case(['--cli-auto-prompt'], 'on-partial', 'on'),
        Case(['--no-cli-auto-prompt'], 'on-partial', 'off'),
        Case(['--version'], 'on', 'off'),
        Case(['help'], 'on', 'off'),
    ]


class FakeSession:
    def __init__(self, emitter=None):
        self.operation = None
        if emitter is None:
            emitter = HierarchicalEmitter()
        self.emitter = emitter
        self.profile = None
        self.stream_logger_args = None
        self.credentials = 'fakecredentials'
        self.session_vars = {}
        self.config_store = self._register_config_store()
        self.user_agent_name = 'aws-cli'
        self.user_agent_version = '100.100.100'

    def _register_config_store(self):
        chain_builder = ConfigChainFactory(session=self)
        config_store = ConfigValueStore(
            mapping=create_botocore_default_config_mapping(chain_builder)
        )
        return config_store

    def register(self, event_name, handler):
        self.emitter.register(event_name, handler)

    def emit(self, event_name, **kwargs):
        return self.emitter.emit(event_name, **kwargs)

    def emit_first_non_none_response(self, event_name, **kwargs):
        responses = self.emitter.emit(event_name, **kwargs)
        for _, response in responses:
            if response is not None:
                return response

    def get_component(self, name):
        if name == 'event_emitter':
            return self.emitter
        if name == 'config_store':
            return self.config_store

    def create_client(self, *args, **kwargs):
        client = mock.Mock()
        client.list_objects.return_value = {}
        client.can_paginate.return_value = False
        return client

    def get_available_services(self):
        return ['s3']

    def get_data(self, name):
        return GET_DATA[name]

    def get_config_variable(self, name):
        if name in GET_VARIABLE:
            return GET_VARIABLE[name]
        return self.session_vars[name]

    def get_service_model(self, name, api_version=None):
        return botocore.model.ServiceModel(MINI_SERVICE, service_name='s3')

    def user_agent(self):
        return 'user_agent'

    def set_stream_logger(self, *args, **kwargs):
        self.stream_logger_args = (args, kwargs)

    def get_credentials(self):
        return self.credentials

    def set_config_variable(self, name, value):
        if name == 'profile':
            self.profile = value
        else:
            self.session_vars[name] = value

    def get_scoped_config(self):
        return GET_VARIABLE.copy()


class FakeCommand(BasicCommand):
    def _run_main(self, args, parsed_globals):
        # We just return success. If this code is reached, it means that
        # all the logic in the __call__ method has sucessfully been run.
        # We subclass it here because the default implementation raises
        # an exception and we don't want that behavior.
        return 0


class FakeCommandVerify(FakeCommand):
    def _run_main(self, args, parsed_globals):
        # Verify passed arguments exist and then return success.
        # This will fail if the expected structure is missing, e.g.
        # if a string is passed in args instead of the expected
        # structure from a custom schema.
        assert args.bar[0]['Name'] == 'test'
        return 0


class FakeCLISessionOrchestrator:
    @property
    def session_id(self):
        return 'mysessionid'


class TestCliDriver:
    def setup_method(self):
        self.session = FakeSession()
        self.session.set_config_variable('cli_auto_prompt', 'off')
        self.session.set_config_variable('cli_help_output', None)
        self.driver = CLIDriver(session=self.session)

    def test_session_can_be_passed_in(self):
        assert self.driver.session == self.session

    def test_paginate_rc(self):
        rc = self.driver.main('s3 list-objects --bucket foo'.split())
        assert rc == 0

    def test_no_profile(self):
        self.driver.main('s3 list-objects --bucket foo'.split())
        assert self.driver.session.profile is None

    def test_profile(self):
        self.driver.main('s3 list-objects --bucket foo --profile foo'.split())
        assert self.driver.session.profile == 'foo'

    def test_region_is_set_for_session(self):
        driver = CLIDriver(session=self.session)
        driver.main('s3 list-objects --bucket foo --region us-east-2'.split())
        assert driver.session.get_config_variable('region') == 'us-east-2'

    @mock.patch('awscli.clidriver.set_stream_logger')
    def test_error_logger(self, set_stream_logger):
        self.driver.main('s3 list-objects --bucket foo --profile foo'.split())
        expected = {'log_level': logging.ERROR, 'logger_name': 'awscli'}
        set_stream_logger.assert_called_with(**expected)

    def test_ctrl_c_is_handled(self):
        fake_client = mock.Mock()
        fake_client.list_objects.side_effect = KeyboardInterrupt
        fake_client.can_paginate.return_value = False
        self.driver.session.create_client = mock.Mock(return_value=fake_client)
        rc = self.driver.main('s3 list-objects --bucket foo'.split())
        assert rc == 130

    def test_error_unicode(self):
        stderr_b = io.BytesIO()
        stderr = io.TextIOWrapper(stderr_b, encoding="UTF-8")
        fake_client = mock.Mock()
        fake_client.list_objects.side_effect = Exception("☃")
        fake_client.can_paginate.return_value = False
        self.driver.session.create_client = mock.Mock(return_value=fake_client)
        with mock.patch("sys.stderr", stderr):
            with mock.patch("locale.getpreferredencoding", lambda: "UTF-8"):
                rc = self.driver.main('s3 list-objects --bucket foo'.split())
        stderr.flush()
        assert rc == 255
        assert stderr_b.getvalue().strip() == "aws: [ERROR]: ☃".encode()

    @pytest.mark.parametrize(
        'env_vars',
        [
            {'AWS_CLI_OUTPUT_ENCODING': 'UTF-8'},
            {'PYTHONUTF8': '1'},
        ],
    )
    def test_error_unicode_env_override(self, env_vars):
        stderr_b = io.BytesIO()
        stderr = io.TextIOWrapper(stderr_b, encoding="cp1252")
        fake_client = mock.Mock()
        fake_client.list_objects.side_effect = Exception("☃")
        fake_client.can_paginate.return_value = False
        self.driver.session.create_client = mock.Mock(return_value=fake_client)
        with mock.patch.dict(os.environ, env_vars):
            with mock.patch("sys.stderr", stderr):
                rc = self.driver.main('s3 list-objects --bucket foo'.split())
        stderr.flush()
        assert rc == 255
        assert stderr_b.getvalue().strip() == "aws: [ERROR]: ☃".encode()

    def test_invalid_output_encoding_throws(self):
        stderr_b = io.BytesIO()
        fake_stderr = io.TextIOWrapper(stderr_b)
        with mock.patch.dict(
            os.environ, {'AWS_CLI_OUTPUT_ENCODING': 'invalid'}
        ):
            with contextlib.redirect_stderr(fake_stderr):
                rc = self.driver.main('s3 list-objects --bucket foo'.split())
        assert rc == 255
        fake_stderr.flush()
        assert (
            stderr_b.getvalue().decode().strip()
            == 'aws: [ERROR]: Unknown codec `invalid` specified for AWS_CLI_OUTPUT_ENCODING.'
        )

    def test_can_access_subcommand_table(self):
        table = self.driver.subcommand_table
        assert list(table) == self.session.get_available_services()

    def test_can_access_argument_table(self):
        arg_table = self.driver.arg_table
        expected = list(GET_DATA['cli']['options'])
        assert list(arg_table) == expected

    def test_cli_driver_can_keep_log_handlers(self):
        fake_stderr = io.StringIO()
        with contextlib.redirect_stderr(fake_stderr):
            driver = create_clidriver(['--debug'])
            rc = driver.main(['ec2', '--debug'])
        assert rc == 252
        assert 2 == fake_stderr.getvalue().count('CLI version:')

    def test_cli_driver_can_remove_log_handlers(self):
        fake_stderr = io.StringIO()
        with contextlib.redirect_stderr(fake_stderr):
            driver = create_clidriver(['--debug'])
            rc = driver.main(['ec2'])
        assert rc == 252
        assert 1 == fake_stderr.getvalue().count('CLI version:')

    @mock.patch('awscrt.io.set_log_level')
    def test_debug_enables_crt_logging(self, mock_init_logging):
        with contextlib.redirect_stderr(io.StringIO()):
            self.driver.main(
                ['s3', 'list-objects', '--bucket', 'foo', '--debug']
            )
        mock_init_logging.assert_called_with(
            awscrt.io.LogLevel.Debug,
        )

    @mock.patch('awscrt.io.set_log_level')
    def test_no_debug_disables_crt_logging(self, mock_init_logging):
        self.driver.main(['s3', 'list-objects', '--bucket', 'foo'])
        mock_init_logging.assert_called_with(
            awscrt.io.LogLevel.NoLogs,
        )

    def test_throw_error_if_both_args_specified(self):
        args = ['--cli-auto-prompt', '--no-cli-auto-prompt']
        with pytest.raises(ParamValidationError):
            validate_auto_prompt_args_are_mutually_exclusive(args)

    @pytest.mark.parametrize('case', _generate_auto_prompt_resolve_cases())
    def test_auto_prompt_resolve_mode(self, case):
        driver = create_clidriver()
        driver.session.set_config_variable(
            'cli_auto_prompt', case.config_variable
        )
        result = resolve_auto_prompt_mode(case.args, driver.session)
        assert result == case.expected_result

    def test_auto_prompt_resolve_mode_on_non_existing_profile(self):
        driver = create_clidriver()
        driver.session.set_config_variable('profile', 'not_exist')
        result = resolve_auto_prompt_mode([], driver.session)
        assert result == 'off'


class TestCliDriverHooks(unittest.TestCase):
    # These tests verify the proper hooks are emitted in clidriver.
    def setUp(self):
        self.session = FakeSession()
        self.session.set_config_variable('cli_auto_prompt', 'off')
        self.session.set_config_variable('cli_help_output', None)
        self.emitter = mock.Mock()
        self.emitter.emit.return_value = []
        self.stdout = StringIO()
        self.stderr = StringIO()
        self.stdout_patch = mock.patch('sys.stdout', self.stdout)
        # self.stdout_patch.start()
        self.stderr_patch = mock.patch('sys.stderr', self.stderr)
        self.stderr_patch.start()

    def tearDown(self):
        # self.stdout_patch.stop()
        self.stderr_patch.stop()

    def assert_events_fired_in_order(self, events):
        args = self.emitter.emit.call_args_list
        actual_events = [arg[0][0] for arg in args]
        self.assertEqual(actual_events, events)

    def serialize_param(self, param, value, **kwargs):
        if kwargs['cli_argument'].name == 'bucket':
            return value + '-altered!'

    def test_expected_events_are_emitted_in_order(self):
        self.emitter.emit.return_value = []
        self.session.emitter = self.emitter
        driver = CLIDriver(session=self.session)
        driver.main('s3 list-objects --bucket foo'.split())
        self.assert_events_fired_in_order(
            [
                # Events fired while parser is being created.
                'building-command-table.main',
                'building-top-level-params',
                'top-level-args-parsed',
                'session-initialized',
                'building-command-table.s3',
                'building-argument-table.s3.list-objects',
                'before-building-argument-table-parser.s3.list-objects',
                'building-command-table.s3_list-objects',
                'operation-args-parsed.s3.list-objects',
                'load-cli-arg.s3.list-objects.bucket',
                'process-cli-arg.s3.list-objects',
                'load-cli-arg.s3.list-objects.marker',
                'load-cli-arg.s3.list-objects.max-keys',
                'calling-command.s3.list-objects',
            ]
        )

    def test_create_help_command(self):
        # When we generate the HTML docs, we don't actually run
        # commands, we just call the create_help_command methods.
        # We want to make sure that in this case, the corresponding
        # building-command-table events are fired.
        # The test above will prove that is true when running a command.
        # This test proves it is true when generating the HTML docs.
        self.emitter.emit.return_value = []
        self.session.emitter = self.emitter
        driver = CLIDriver(session=self.session)
        main_hc = driver.create_help_command()
        command = main_hc.command_table['s3']
        command.create_help_command()
        self.assert_events_fired_in_order(
            [
                # Events fired while parser is being created.
                'building-command-table.main',
                'building-top-level-params',
                'building-command-table.s3',
            ]
        )

    def test_unknown_command_suggests_help(self):
        driver = CLIDriver(
            session=self.session,
            error_handler=construct_cli_error_handlers_chain(),
        )

        # Note the typo in 'list-objects'
        rc = driver.main(
            's3 list-objecst --bucket foo --unknown-arg foo'.split()
        )
        self.assertEqual(rc, 252)
        # Tell the user what went wrong.
        self.assertIn(
            "Found invalid choice 'list-objecst'", self.stderr.getvalue()
        )
        # Offer the user a suggestion.
        self.assertIn(
            "Maybe you meant:\n\n  * list-objects", self.stderr.getvalue()
        )


class TestSearchPath(unittest.TestCase):
    @mock.patch('os.pathsep', ';')
    @mock.patch('os.environ', {'AWS_DATA_PATH': 'c:\\foo;c:\\bar'})
    def test_windows_style_search_path(self):
        driver = CLIDriver()
        # Our two overrides should be the last two elements in the search path.
        search_paths = driver.session.get_component('data_loader').search_paths
        self.assertIn('c:\\foo', search_paths)
        self.assertIn('c:\\bar', search_paths)


class TestAWSCommand(BaseAWSCommandParamsTest):
    # These tests will simulate running actual aws commands
    # but with the http part mocked out.
    def setUp(self):
        super().setUp()
        self.stderr = StringIO()
        self.stderr_patch = mock.patch('sys.stderr', self.stderr)
        self.stderr_patch.start()

    def tearDown(self):
        super().tearDown()
        self.stderr_patch.stop()

    def inject_new_param(self, argument_table, **kwargs):
        argument = CustomArgument('unknown-arg', {})
        argument.add_to_arg_table(argument_table)

    def inject_command(self, command_table, session, **kwargs):
        command = FakeCommand(session)
        command.NAME = 'foo'
        command.ARG_TABLE = [{'name': 'bar', 'action': 'store'}]
        command_table['foo'] = command

    def inject_command_schema(self, command_table, session, **kwargs):
        command = FakeCommandVerify(session)
        command.NAME = 'foo'

        # Build a schema using all the types we are interested in
        schema = {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "Name": {"type": "string", "required": True},
                    "Count": {"type": "integer"},
                },
            },
        }

        command.ARG_TABLE = [{'name': 'bar', 'schema': schema}]

        command_table['foo'] = command

    def test_event_emission_for_top_level_params(self):
        driver = create_clidriver()
        # --unknown-foo is an unknown arg, so we expect a 252 rc.
        rc = driver.main('ec2 describe-instances --unknown-arg foo'.split())
        self.assertEqual(rc, 252)
        self.assertIn('Unknown options: --unknown-arg', self.stderr.getvalue())

        # The argument table is memoized in the CLIDriver object. So
        # when we call main() above, it will get created and cached
        # and the argument table won't get created again (and therefore
        # the building-top-level-params event will not get generated again).
        # So, for this test we need to create a new driver object.
        driver = create_clidriver()
        driver.session.register(
            'building-top-level-params', self.inject_new_param
        )
        driver.session.register(
            'top-level-args-parsed',
            lambda parsed_args, **kwargs: args_seen.append(parsed_args),
        )

        args_seen = []

        # Now we should get an rc of 0 as the arg is expected
        # (though nothing actually does anything with the arg).
        self.patch_make_request()
        rc = driver.main('ec2 describe-instances --unknown-arg foo'.split())
        self.assertEqual(rc, 0)
        self.assertEqual(len(args_seen), 1)
        self.assertEqual(args_seen[0].unknown_arg, 'foo')

    @mock.patch('awscli.paramfile.URIArgumentHandler', spec=URIArgumentHandler)
    def test_custom_arg_paramfile(self, mock_handler):
        mock_paramfile = mock.Mock(autospec=True)
        mock_paramfile.return_value = None
        mock_handler.return_value = mock_paramfile

        driver = create_clidriver()
        driver.session.register(
            'building-argument-table', self.inject_new_param
        )

        self.patch_make_request()
        rc = driver.main(
            'ec2 describe-instances --unknown-arg file:///foo'.split()
        )

        self.assertEqual(rc, 0)

        # Make sure uri_param was called
        mock_paramfile.assert_any_call(
            event_name='load-cli-arg.ec2.describe-instances.unknown-arg',
            operation_name='describe-instances',
            param=mock.ANY,
            service_name='ec2',
            value='file:///foo',
        )
        # Make sure it was called with our passed-in URI
        self.assertEqual(
            'file:///foo', mock_paramfile.call_args_list[-1][1]['value']
        )

    @mock.patch('awscli.paramfile.URIArgumentHandler', spec=URIArgumentHandler)
    def test_custom_command_paramfile(self, mock_handler):
        mock_paramfile = mock.Mock(autospec=True)
        mock_paramfile.return_value = None
        mock_handler.return_value = mock_paramfile

        driver = create_clidriver()
        driver.session.register('building-command-table', self.inject_command)

        self.patch_make_request()
        rc = driver.main('ec2 foo --bar file:///foo'.split())

        self.assertEqual(rc, 0)

        mock_paramfile.assert_any_call(
            event_name='load-cli-arg.custom.foo.bar',
            operation_name='foo',
            param=mock.ANY,
            service_name='custom',
            value='file:///foo',
        )

    def test_custom_command_schema(self):
        driver = create_clidriver()
        driver.session.register(
            'building-command-table', self.inject_command_schema
        )

        self.patch_make_request()

        # Test single shorthand item
        rc = driver.main('ec2 foo --bar Name=test,Count=4'.split())

        self.assertEqual(rc, 0)

        # Test shorthand list of items with optional values
        rc = driver.main(
            'ec2 foo --bar Name=test,Count=4 Name=another'.split()
        )

        self.assertEqual(rc, 0)

        # Test missing require shorthand item
        rc = driver.main('ec2 foo --bar Count=4'.split())

        self.assertEqual(rc, 252)

        # Test extra unknown shorthand item
        rc = driver.main('ec2 foo --bar Name=test,Unknown='.split())

        self.assertEqual(rc, 252)

        # Test long form JSON
        rc = driver.main('ec2 foo --bar {"Name":"test","Count":4}'.split())

        self.assertEqual(rc, 0)

        # Test malformed long form JSON
        rc = driver.main('ec2 foo --bar {"Name":"test",Count:4}'.split())

        self.assertEqual(rc, 252)

    def test_empty_params_gracefully_handled(self):
        # Simulates the equivalent in bash: --identifies ""
        cmd = 'ses get-identity-dkim-attributes --identities'.split()
        cmd.append('')
        self.assert_params_for_cmd(cmd, expected_rc=0)

    def test_file_param_does_not_exist(self):
        driver = create_clidriver()
        rc = driver.main(
            'ec2 describe-instances '
            '--filters file://does/not/exist.json'.split()
        )
        self.assertEqual(rc, 252)
        error_msg = self.stderr.getvalue()
        self.assertIn(
            "Error parsing parameter '--filters': "
            "Unable to load paramfile file://does/not/exist.json",
            error_msg,
        )
        self.assertIn("No such file or directory", error_msg)

    def test_aws_configure_in_error_message_no_credentials(self):
        driver = create_clidriver()

        def raise_exception(*args, **kwargs):
            raise NoCredentialsError()

        driver.session.register(
            'building-command-table',
            lambda command_table, **kwargs: command_table.__setitem__(
                'ec2', raise_exception
            ),
        )
        with mock.patch('sys.stderr') as f:
            driver.main('ec2 describe-instances'.split())
        self.assertEqual(
            f.write.call_args_list[1][0][0],
            'aws: [ERROR]: An error occurred (NoCredentials): '
            'Unable to locate credentials. '
            'You can configure credentials by running "aws login".',
        )

    def test_override_calling_command(self):
        self.driver = create_clidriver()

        # Make a function that will return an override such that its value
        # is used over whatever is returned by the invoker which is usually
        # zero.
        def override_with_rc(**kwargs):
            return 20

        self.driver.session.register('calling-command', override_with_rc)
        rc = self.driver.main('ec2 describe-instances'.split())
        # Check that the overriden rc is as expected.
        self.assertEqual(rc, 20)

    def test_override_calling_command_error(self):
        self.driver = create_clidriver()

        # Make a function that will return an error. The handler will cause
        # an error to be returned and later raised.
        def override_with_error(**kwargs):
            return ValueError()

        self.driver.session.register('calling-command', override_with_error)
        # An exception should be thrown as a result of the handler, which
        # will result in 255 rc.
        rc = self.driver.main('ec2 describe-instances'.split())
        self.assertEqual(rc, 255)

    def test_help_blurb_in_error_message(self):
        rc = self.driver.main([])
        self.assertEqual(rc, 252)
        self.assertIn(HELP_BLURB, self.stderr.getvalue())

    def test_help_blurb_in_service_error_message(self):
        rc = self.driver.main(['ec2'])
        self.assertEqual(rc, 252)
        self.assertIn(HELP_BLURB, self.stderr.getvalue())

    def test_help_blurb_in_operation_error_message(self):
        rc = self.driver.main(['s3api', 'list-objects'])
        self.assertEqual(rc, 252)
        self.assertIn(HELP_BLURB, self.stderr.getvalue())

    def test_help_blurb_in_unknown_argument_error_message(self):
        args = ['s3api', 'list-objects', '--help']
        driver = create_clidriver(args)
        rc = driver.main(args)
        self.assertEqual(rc, 252)
        self.assertIn(HELP_BLURB, self.stderr.getvalue())

    def test_idempotency_token_is_not_required_in_help_text(self):
        rc = self.driver.main(['servicecatalog', 'create-constraint'])
        self.assertEqual(rc, 252)
        self.assertNotIn('--idempotency-token', self.stderr.getvalue())

    @mock.patch(
        'awscli.telemetry._get_cli_session_orchestrator',
        return_value=FakeCLISessionOrchestrator(),
    )
    @mock.patch('awscli.clidriver.platform.system', return_value='Linux')
    @mock.patch('awscli.clidriver.platform.machine', return_value='x86_64')
    @mock.patch('awscli.clidriver.distro.id', return_value='amzn')
    @mock.patch('awscli.clidriver.distro.major_version', return_value='1')
    def test_user_agent_for_linux(self, *args):
        driver = create_clidriver()
        expected_user_agent = (
            'md/installer#source md/distrib#amzn.1 sid/mysessionid'
        )
        self.assertEqual(expected_user_agent, driver.session.user_agent_extra)

    def test_user_agent(self, *args):
        driver = create_clidriver()
        user_agent_extra_pattern = re.compile(r'^md/installer#source')
        self.assertIsNotNone(
            user_agent_extra_pattern.match(driver.session.user_agent_extra)
        )
        # check that distro didn't fail
        self.assertFalse('unknown' in driver.session.user_agent_extra)

    @mock.patch('awscli.clidriver.platform.system', return_value='Linux')
    @mock.patch('awscli.clidriver.distro.id', side_effect=Exception())
    def test_user_agent_handles_distro_exception(self, *args):
        driver = create_clidriver()
        self.assertTrue('unknown' in driver.session.user_agent_extra)


class TestHowClientIsCreated(BaseAWSCommandParamsTest):
    def setUp(self):
        super().setUp()
        self.endpoint_creator_patch = mock.patch(
            'botocore.args.EndpointCreator'
        )
        self.endpoint_creator = self.endpoint_creator_patch.start()
        self.create_endpoint = (
            self.endpoint_creator.return_value.create_endpoint
        )
        self.endpoint = self.create_endpoint.return_value
        self.endpoint.host = 'https://example.com'
        # Have the endpoint give a dummy empty response.
        http_response = AWSResponse(None, 200, {}, None)
        self.endpoint.make_request.return_value = (http_response, {})

    def tearDown(self):
        super().tearDown()
        self.endpoint_creator_patch.stop()

    def test_aws_with_endpoint_url(self):
        self.assert_params_for_cmd(
            'ec2 describe-instances --endpoint-url https://foobar.com/',
            expected_rc=0,
        )
        self.assertEqual(
            self.create_endpoint.call_args[1]['endpoint_url'],
            'https://foobar.com/',
        )

    def test_aws_with_region(self):
        self.assert_params_for_cmd(
            'ec2 describe-instances --region us-west-2', expected_rc=0
        )
        self.assertEqual(
            self.create_endpoint.call_args[1]['region_name'], 'us-west-2'
        )

    def test_can_use_new_aws_region_env_var(self):
        self.environ['AWS_REGION'] = 'us-east-2'
        self.environ['AWS_DEFAULT_REGION'] = 'us-west-1'
        self.assert_params_for_cmd('ec2 describe-instances', expected_rc=0)
        self.assertEqual(
            self.create_endpoint.call_args[1]['region_name'], 'us-east-2'
        )

    def test_aws_with_verify_false(self):
        self.assert_params_for_cmd(
            'ec2 describe-instances --region us-east-1 --no-verify-ssl',
            expected_rc=0,
        )
        # Because we used --no-verify-ssl, create_endpoint should be
        # called with verify=False
        call_args = self.create_endpoint.call_args
        self.assertFalse(call_args[1]['verify'])

    def test_aws_with_cacert_env_var(self):
        self.environ['AWS_CA_BUNDLE'] = '/path/cacert.pem'
        self.assert_params_for_cmd(
            'ec2 describe-instances --region us-east-1', expected_rc=0
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['verify'], '/path/cacert.pem')

    def test_aws_with_read_timeout(self):
        self.assert_params_for_cmd(
            'lambda invoke --function-name foo out.log --cli-read-timeout 90',
            expected_rc=0,
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['timeout'][1], 90)

    def test_aws_with_blocking_read_timeout(self):
        self.assert_params_for_cmd(
            'lambda invoke --function-name foo out.log --cli-read-timeout 0',
            expected_rc=0,
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['timeout'][1], None)

    def test_aws_with_connnect_timeout(self):
        self.assert_params_for_cmd(
            'lambda invoke --function-name foo out.log '
            '--cli-connect-timeout 90',
            expected_rc=0,
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['timeout'][0], 90)

    def test_aws_with_blocking_connect_timeout(self):
        self.assert_params_for_cmd(
            'lambda invoke --function-name foo out.log '
            '--cli-connect-timeout 0',
            expected_rc=0,
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['timeout'][0], None)

    def test_aws_with_read_and_connnect_timeout(self):
        self.assert_params_for_cmd(
            'lambda invoke --function-name foo out.log '
            '--cli-read-timeout 70 --cli-connect-timeout 90',
            expected_rc=0,
        )
        call_args = self.create_endpoint.call_args
        self.assertEqual(call_args[1]['timeout'], (90, 70))


class TestVerifyArgument(BaseAWSCommandParamsTest):
    def setUp(self):
        super().setUp()
        self.driver.session.register('top-level-args-parsed', self.record_args)
        self.recorded_args = None

    def record_args(self, parsed_args, **kwargs):
        self.recorded_args = parsed_args

    def test_no_verify_argument(self):
        self.assert_params_for_cmd(
            's3api list-buckets --no-verify-ssl'.split()
        )
        self.assertFalse(self.recorded_args.verify_ssl)

    def test_verify_argument_is_none_by_default(self):
        self.assert_params_for_cmd('s3api list-buckets'.split())
        self.assertIsNone(self.recorded_args.verify_ssl)


class TestFormatter(BaseAWSCommandParamsTest):
    def test_bad_output(self):
        with self.assertRaises(ValueError):
            formatter.get_formatter('bad-type', None)


class TestCLICommand(unittest.TestCase):
    def setUp(self):
        self.cmd = CLICommand()

    def test_name(self):
        with self.assertRaises(NotImplementedError):
            self.cmd.name
        with self.assertRaises(NotImplementedError):
            self.cmd.name = 'foo'

    def test_lineage(self):
        self.assertEqual(self.cmd.lineage, [self.cmd])

    def test_lineage_names(self):
        with self.assertRaises(NotImplementedError):
            self.cmd.lineage_names

    def test_arg_table(self):
        self.assertEqual(self.cmd.arg_table, {})


class TestServiceCommand(unittest.TestCase):
    def setUp(self):
        self.name = 'foo'
        self.session = FakeSession()
        self.session.set_config_variable('cli_help_output', None)
        self.cmd = ServiceCommand(self.name, self.session)

    def test_can_access_subcommand_table(self):
        table = self.cmd.subcommand_table
        expected = [
            xform_name(op, '-')
            for op in self.session.get_service_model(self.name).operation_names
        ]
        self.assertEqual(set(table), set(expected))

    def test_can_access_arg_table(self):
        self.assertEqual(self.cmd.arg_table, {})

    def test_name(self):
        self.assertEqual(self.cmd.name, self.name)
        self.cmd.name = 'bar'
        self.assertEqual(self.cmd.name, 'bar')

    def test_lineage(self):
        cmd = CLICommand()
        self.assertEqual(self.cmd.lineage, [self.cmd])
        self.cmd.lineage = [cmd]
        self.assertEqual(self.cmd.lineage, [cmd])

    def test_lineage_names(self):
        self.assertEqual(self.cmd.lineage_names, ['foo'])

    def test_pass_lineage_to_child(self):
        # In order to introspect the service command's subcommands
        # we introspect the subcommand via the help command since
        # a service command's command table is not public.
        help_command = self.cmd.create_help_command()
        child_cmd = help_command.command_table['list-objects']
        self.assertEqual(child_cmd.lineage, [self.cmd, child_cmd])
        self.assertEqual(child_cmd.lineage_names, ['foo', 'list-objects'])

    def test_help_event_class(self):
        # Ensures it sends the right event name to the help command
        help_command = self.cmd.create_help_command()
        self.assertEqual(help_command.event_class, 'foo')
        child_cmd = help_command.command_table['list-objects']
        # Check the ``ServiceOperation`` class help command as well
        child_help_cmd = child_cmd.create_help_command()
        self.assertEqual(child_help_cmd.event_class, 'foo.list-objects')


class TestServiceOperation(unittest.TestCase):
    def setUp(self):
        self.name = 'list-objects'
        self.session = FakeSession()
        operation = self.session.get_service_model('s3').operation_model(
            'ListObjects'
        )
        self.mock_operation = operation
        self.cmd = ServiceOperation(
            self.name, None, None, operation, self.session
        )

    def test_can_access_subcommand_table(self):
        self.assertEqual(self.cmd.subcommand_table, {})

    def test_can_access_arg_table(self):
        self.assertEqual(
            set(self.cmd.arg_table), set(['bucket', 'marker', 'max-keys'])
        )

    def test_name(self):
        self.assertEqual(self.cmd.name, self.name)
        self.cmd.name = 'bar'
        self.assertEqual(self.cmd.name, 'bar')

    def test_lineage(self):
        cmd = CLICommand()
        self.assertEqual(self.cmd.lineage, [self.cmd])
        self.cmd.lineage = [cmd]
        self.assertEqual(self.cmd.lineage, [cmd])

    def test_lineage_names(self):
        self.assertEqual(self.cmd.lineage_names, ['list-objects'])

    def test_deprecated_operation(self):
        self.mock_operation.deprecated = True
        cmd = ServiceOperation(
            self.name, None, None, self.mock_operation, None
        )
        self.assertTrue(getattr(cmd, '_UNDOCUMENTED'))

    def test_idempotency_token_is_not_required(self):
        session = FakeSession()
        name = 'IdempotentOperation'
        service_model = session.get_service_model('s3')
        operation_model = service_model.operation_model(name)
        cmd = ServiceOperation(name, None, None, operation_model, session)
        arg_table = cmd.arg_table
        token_argument = arg_table.get('token')
        self.assertFalse(
            token_argument.required,
            'Idempotency tokens should not be required',
        )


class TestAWSCLIEntryPoint(unittest.TestCase):
    def setUp(self):
        self.driver = mock.Mock()

        def _create_fake_cli_driver(*args):
            self.driver.session.user_agent_extra = ''
            return self.driver

        self.prompt_patch = mock.patch(
            'awscli.autoprompt.core.AutoPromptDriver'
        )
        self.crete_driver_patch = mock.patch(
            'awscli.clidriver.create_clidriver'
        )
        self.resolve_mode_patch = mock.patch(
            'awscli.clidriver.resolve_auto_prompt_mode'
        )
        prompt_driver_class = self.prompt_patch.start()
        self.create_clidriver = self.crete_driver_patch.start()
        self.resolve_auto_prompt_mode = self.resolve_mode_patch.start()
        self.create_clidriver.side_effect = _create_fake_cli_driver
        self.prompt_driver = mock.Mock()
        prompt_driver_class.return_value = self.prompt_driver

    def tearDown(self):
        self.prompt_patch.stop()
        self.crete_driver_patch.stop()
        self.resolve_mode_patch.stop()

    def test_recreate_driver_in_partial_mode_on_param_err(self):
        self.resolve_auto_prompt_mode.return_value = 'on-partial'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        rc = entry_point.main([])
        self.assertEqual(self.create_clidriver.call_count, 2)
        self.assertEqual(rc, 252)

    def test_not_recreate_driver_in_partial_mode_on_success(self):
        self.resolve_auto_prompt_mode.return_value = 'on-partial'
        self.driver.main.return_value = 0
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        rc = entry_point.main([])
        self.assertEqual(self.create_clidriver.call_count, 1)
        self.assertEqual(rc, 0)

    def test_not_recreate_driver_in_on_mode(self):
        self.resolve_auto_prompt_mode.return_value = 'on'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        rc = entry_point.main([])
        self.assertEqual(self.create_clidriver.call_count, 1)
        self.assertEqual(rc, 252)

    def test_not_recreate_driver_in_off_mode(self):
        self.resolve_auto_prompt_mode.return_value = 'off'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        rc = entry_point.main([])
        self.assertEqual(self.create_clidriver.call_count, 1)
        self.assertEqual(rc, 252)

    def test_handle_exception_in_main(self):
        self.resolve_auto_prompt_mode.return_value = 'on'
        self.prompt_driver.prompt_for_args.side_effect = Exception('error')
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        fake_stderr = io.StringIO()
        with contextlib.redirect_stderr(fake_stderr):
            rc = entry_point.main([])
        self.assertEqual(rc, 255)
        self.assertIn('error', fake_stderr.getvalue())

    def test_update_user_agent_in_on_mode(self):
        self.resolve_auto_prompt_mode.return_value = 'on'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        entry_point.main([])
        self.assertEqual(self.driver.session.user_agent_extra, 'md/prompt#on')

    def test_not_update_user_agent_in_off_mode(self):
        self.resolve_auto_prompt_mode.return_value = 'off'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        entry_point.main([])
        self.assertEqual(self.driver.session.user_agent_extra, 'md/prompt#off')

    def test_update_user_agent_in_partial_mode_on_param_err(self):
        self.resolve_auto_prompt_mode.return_value = 'on-partial'
        self.driver.main.return_value = 252
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        entry_point.main([])
        self.assertEqual(
            self.driver.session.user_agent_extra, 'md/prompt#partial'
        )

    def test_not_update_user_agent_in_partial_mode_on_success(self):
        self.resolve_auto_prompt_mode.return_value = 'on-partial'
        self.driver.main.return_value = 0
        entry_point = awscli.clidriver.AWSCLIEntryPoint()
        entry_point.main([])
        self.assertEqual(self.driver.session.user_agent_extra, 'md/prompt#off')


class TextCreateCLIDriver(unittest.TestCase):
    def test_create_cli_driver_parse_args(self):
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr):
            create_clidriver(['--debug'])
        self.assertIn('CLI version', stderr.getvalue())

    def test_create_cli_driver_wo_args(self):
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr):
            create_clidriver()
        self.assertIn('', stderr.getvalue())


if __name__ == '__main__':
    unittest.main()
